# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output routines for gravitational-wave frame (GWF) format files.

The GWF format is defined in :dcc:`LIGO-T970130`.

The functions in this module are the ones exposed to the user, but
connect to the backend-specific functions in the other modules in
the same subpackage.
"""

from __future__ import annotations

import warnings
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

from ....io import (
    cache as io_cache,
    gwf as io_gwf,
)
from ....io.utils import (
    FileLike,
    file_path,
)
from ....segments import Segment
from ....time import to_gps
from ....utils.decorators import deprecated_function
from ... import (
    StateVector,
    StateVectorDict,
    TimeSeries,
    TimeSeriesDict,
)
from .utils import _channel_dict_kwarg

if TYPE_CHECKING:
    from typing import IO

    from ....time import SupportsToGps
    from ... import (
        Bits,
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# preferentially ordered list of GWF backends
BACKENDS: list[str] = [
    "frameCPP",
    "FrameL",
    "LALFrame",
]


# -- read ----------------------------

def read_timeseriesdict(
    source: str | list[str],
    channels: list[str],
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    scaled: bool | None = None,
    type: str | dict[str, str] | None = None,
    backend: str | None = None,
    series_class: type[TimeSeriesBase] = TimeSeries,
) -> TimeSeriesBaseDict:
    """Read the data for a list of channels from a GWF data source.

    Parameters
    ----------
    source : `str`, `Path`, `file`, or `list`
        Source of data, any of the following:

        - path to a single GWF file
        - path to a cache file
        - `list` of GWF file paths.

    channels : `list`
        List of data channel names (or `Channel` objects) to read from GWF.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data.

    type : `dict`, optional
        A `dict` of ``(name, channel-type)`` pairs, where ``channel-type``
        can be one of ``'adc'``, ``'proc'``, or ``'sim'``.

    backend : `str`, optional
        GWF backend to use when reading.
        Default is the value of the ``GWPY_FRAME_LIBRARY`` environment
        variable, or 'frameCPP'.

    series_class : `type[TimeSeriesBase]`, optional
        The type to use for all series objects.
        Default is the class being used for `.read()`.

    Returns
    -------
    dict : :class:`~gwpy.timeseries.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs
    """
    read_func = io_gwf.get_backend_function(
        "read",
        backend=backend,
        backends=BACKENDS,
        package=__package__,
    )

    # -- from here read data

    if start:
        start = to_gps(start)
    if end:
        end = to_gps(end)

    # read cache file up-front
    if (
        (isinstance(source, str) and source.endswith((".lcf", ".cache")))
        or (
            isinstance(source, FileLike)
            and hasattr(source, "name")
            and source.name.endswith((".lcf", ".cache"))
        )
    ):
        source = io_cache.read_cache(source)
    # separate cache into contiguous segments
    if io_cache.is_cache(source):
        if start is not None and end is not None:
            source = io_cache.sieve(
                source,
                segment=Segment(start, end),
            )
        source = list(io_cache.find_contiguous(source))
    # convert everything else into a list if needed
    if not isinstance(source, list | tuple):
        source = [source]

    # now read the data
    return read_func(
        source,
        channels,
        start=start,
        end=end,
        scaled=scaled,
        type=type,
        series_class=series_class,
    )


def read_timeseries(
    source: str | list[str],
    channel: str,
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    scaled: bool | None = None,
    type: str | dict[str, str] | None = None,
    backend: str | None = None,
    series_class: type[TimeSeriesBase] = TimeSeries,
) -> TimeSeries | StateVector:
    """Read a `TimeSeriesa from one or more GWF files.

    Parameters
    ----------
    source : `str`, `Path`, `file`, or `list`
        Source of data, any of the following:

        - path to a single GWF file
        - path to a cache file
        - `list` of GWF file paths.

    channel : `str`, `~gwpy.detector.Channel`
        Name of data channel to read from GWF.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data.

    type : `dict`, optional
        A `dict` of ``(name, channel-type)`` pairs, where ``channel-type``
        can be one of ``'adc'``, ``'proc'``, or ``'sim'``.

    backend : `str`, optional
        GWF backend to use when reading.
        Default is 'frameCPP'.

    series_class : `type[TimeSeriesBase]`, optional
        The type to use for all series objects.
        Default is the class being used for `.read()`.

    Returns
    -------
    series : `~gwpy.timeseries.TimeSeries`
        A `TimeSeries` containing the data read from GWF.

    Raises
    ------
    ValueError
        If the given channel name isn't found, or the ``[start, end)``
        interval doesn't match the data available from the GWF source.
    """
    return read_timeseriesdict(
        source,
        [channel],
        start=start,
        end=end,
        scaled=scaled,
        type=type,
        backend=backend,
        series_class=series_class,
    )[channel]


def read_statevector(
    source: str | list[str],
    channel: str,
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    bits: list[str] | Bits | None = None,
    scaled: bool | None = None,
    type: str | dict[str, str] | None = None,
    backend: str | None = None,
    series_class: type[StateVector] = StateVector,
) -> StateVector:
    """Read a `StateVector` from one or more GWF files.

    Parameters
    ----------
    source : `str`, `Path`, `file`, or `list`
        Source of data, any of the following:

        - path to a single GWF file
        - path to a cache file
        - `list` of GWF file paths.

    channel : `str`, `~gwpy.detector.Channel`
        Name of data channel to read from GWF.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    bits : `Bits`, `list`, optional
        List of bits defining this `StateVector`.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data.

    type : `dict`, optional
        A `dict` of ``(name, channel-type)`` pairs, where ``channel-type``
        can be one of ``'adc'``, ``'proc'``, or ``'sim'``.

    backend : `str`, optional
        GWF backend to use when reading.
        Default is 'frameCPP'.

    series_class : `type[TimeSeriesBase]`, optional
        The type to use for all series objects.
        Default is the class being used for `.read()`.

    Returns
    -------
    statevector : `~gwpy.timeseries.StateVector`
        A `StateVector` containing the data read from GWF.

    Raises
    ------
    ValueError
        If the given channel name isn't found, or the ``[start, end)``
        interval doesn't match the data available from the GWF source.
    """
    statevector = read_timeseries(
        source,
        channel,
        start=start,
        end=end,
        scaled=scaled,
        type=type,
        backend=backend,
        series_class=series_class,
    )
    statevector.bits = bits
    return statevector


def read_statevectordict(
    source: str | list[str],
    channels: list[str],
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    bits: list[str] | Bits | None = None,
    scaled: bool | None = None,
    type: str | dict[str, str] | None = None,
    backend: str | None = None,
    series_class: type[StateVector] = StateVector,
) -> StateVectorDict:
    """Read a `StateVectorDict` from GWF files.

    Parameters
    ----------
    source : `str`, `Path`, `file`, or `list`
        Source of data, any of the following:

        - path to a single GWF file
        - path to a cache file
        - `list` of GWF file paths.

    channels : `list`
        List of data channel names (or `Channel` objects) to read from GWF.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    bits : `dict` of ``(str, bits)`` pairs, `list`
        A `dict` giving the bit definition for each channel, or a `list`
        given the bits to assign identically to all channels.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data.

    type : `dict`, optional
        A `dict` of ``(name, channel-type)`` pairs, where ``channel-type``
        can be one of ``'adc'``, ``'proc'``, or ``'sim'``.

    backend : `str`, optional
        GWF backend to use when reading.
        Default is the value of the ``GWPY_FRAME_LIBRARY`` environment
        variable, or 'frameCPP'.

    series_class : `type[TimeSeriesBase]`, optional
        The type to use for all series objects.
        Default is the class being used for `.read()`.

    Returns
    -------
    dict : :class:`~gwpy.timeseries.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs
    """
    # read data
    svd = StateVectorDict(read_timeseriesdict(
        source,
        channels,
        start=start,
        end=end,
        scaled=scaled,
        type=type,
        backend=backend,
        series_class=series_class,
    ))

    # add bits definitions
    bitss = _channel_dict_kwarg(
        bits,
        channels,
        list,
        varname="bits",
    )
    for (channel, bits_) in bitss.items():
        svd[channel].bits = bits_
    return svd


# -- write ----------------------------------

def write_timeseriesdict(
    seriesdict: TimeSeriesBaseDict,
    target: str | Path | IO,
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    type: str | None = None,
    name: str | None = None,
    run: int = 0,
    compression: str | int | None = None,
    compression_level: int | None = None,
    backend: str | None = None,
    *,
    overwrite: bool = True,
    append: bool = False,
) -> None:
    """Write a `TimeSeriesDict` in GWF format.

    Parameters
    ----------
    seriesdict : `TimeSeriesDict`
        The data to write.

    target : `str`, `pathlib.Path`, `file`
        The target file or path to write to.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        The GPS start time (``GTime``) of the output ``FrameH`` structure.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        The GPS end time of the output ``FrameH``.
        Used with ``start`` to set the ``Dt`` attribute.

    type : `str`, optional
        The type of the channel, one of 'adc', 'proc', 'sim'.
        Default is 'proc' unless stored in the ``_ctype`` attribute
        of the channel structure.

    name : `str`, optional
        The name of the frame.
        Default is ``"gwpy"`` when using FrameCPP and LALFrame backends,
        or the name of the first channel in the dict when writing with
        FrameL.

        This keyword is not supported when writing with FrameL.

    run : `int`, optional
        The ``FrameH`` run number.
        Values less than ``0`` are reserved for simulated data, should be a
        monotonically increasing integer for experimental runs.

        This keyword is not supported when writing with FrameL.

    compression : `int`, `str`, optional
        Name of compresion algorithm to use, or its endian-appropriate
        ID. One of:

        - ``'RAW'``
        - ``'GZIP'`` (default)
        - ``'DIFF_GZIP'``
        - ``'ZERO_SUPPRESS'``
        - ``'ZERO_SUPPRESS_OTHERWISE_GZIP'``

        Only ``'GZIP'`` is supported when writing with FrameL.

    compression_level : `int`, optional
        Compression level for given method, default is ``6`` for GZIP-based
        methods, otherwise ``0``.

    backend : `str`, optional
        GWF backend to use when reading.
        Default is 'frameCPP'.

    overwrite : `bool`, optional
        If `True` (default), overwrite existing files.

    append : `bool`, optional
        DO NOT USE. This option is not supported for writing GWF files;
        the entire file will be overwritten.
        This option is only present for API consistency with other
        GWpy I/O functions.

    Raises
    ------
    OSError
        If the target file exists and neither ``overwrite`` nor ``append``
        is `True`.
    """
    try:
        target_path = Path(file_path(target))
    except ValueError:
        # not a file path, so cannot check for existence, that's fine
        pass
    else:
        _exists = target_path.exists()
        if _exists and not overwrite:
            msg = f"File exists: {target_path}"
            raise OSError(msg)
        if _exists and append:
            warnings.warn(
                "append=True is not supported for writing GWF files; "
                "the entire file will be overwritten.",
                category=UserWarning,
                stacklevel=2,
            )

    # get backend function
    write_func = io_gwf.get_backend_function(
        "write",
        backend=backend,
        backends=BACKENDS,
        package=__package__,
    )

    # pre-format GPS times
    span = seriesdict.span
    if start is None:
        startgps = span[0]
    else:
        startgps = to_gps(start)
    if end is None:
        endgps = span[1]
    else:
        endgps = to_gps(end)

    write_func(
        seriesdict,
        target,
        startgps,
        endgps,
        type=type,
        name=name,
        run=run,
        compression=compression,
        compression_level=compression_level,
    )


def write_timeseries(
    series: TimeSeriesBase,
    target: str | Path | IO,
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    type: str | None = None,
    name: str | None = None,
    run: int = 0,
    compression: str | int | None = None,
    compression_level: int | None = None,
    backend: str | None = None,
    *,
    overwrite: bool = True,
    append: bool = False,
) -> None:
    """Write a `TimeSeries` to disk in GWF format.

    Parameters
    ----------
    series : `TimeSeries`
        The data to write.

    target : `str`, `pathlib.Path`, `file`
        The target file or path to write to.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        The GPS start time (``GTime``) of the output ``FrameH`` structure.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        The GPS end time of the output ``FrameH``.
        Used with ``start`` to set the ``Dt`` attribute.

    type : `str`, optional
        The type of the channel, one of 'adc', 'proc', 'sim'.
        Default is 'proc' unless stored in the ``_ctype`` attribute
        of the channel structure.

    name : `str`, optional
        The name of the frame.
        Default is ``"gwpy"`` when using FrameCPP and LALFrame backends,
        or the name of the first channel in the dict when writing with
        FrameL.

        This keyword is not supported when writing with FrameL.

    run : `int`, optional
        The ``FrameH`` run number.
        Values less than ``0`` are reserved for simulated data, should be a
        monotonically increasing integer for experimental runs.

        This keyword is not supported when writing with FrameL.

    compression : `int`, `str`, optional
        Name of compresion algorithm to use, or its endian-appropriate
        ID. One of:

        - ``'RAW'``
        - ``'GZIP'`` (default)
        - ``'DIFF_GZIP'``
        - ``'ZERO_SUPPRESS'``
        - ``'ZERO_SUPPRESS_OTHERWISE_GZIP'``

        Only ``'GZIP'`` is supported when writing with FrameL.

    compression_level : `int`, optional
        Compression level for given method, default is ``6`` for GZIP-based
        methods, otherwise ``0``.


    backend : `str`, optional
        GWF backend to use when reading.
        Default is 'frameCPP'.

    overwrite : `bool`, optional
        If `True` (default), overwrite existing files.

    append : `bool`, optional
        DO NOT USE. This option is not supported for writing GWF files;
        the entire file will be overwritten.
        This option is only present for API consistency with other
        GWpy I/O functions.

    Raises
    ------
    OSError
        If the target file exists and neither ``overwrite`` nor ``append``
        is `True`.
    """
    write_timeseriesdict(
        series.DictClass({None: series}),
        target=target,
        start=start,
        end=end,
        type=type,
        name=name,
        run=run,
        compression=compression,
        compression_level=compression_level,
        backend=backend,
        overwrite=overwrite,
        append=append,
    )


# -- register -------------------------------

def register_gwf_backend(backend: str | None) -> None:
    """Register a full set of GWF I/O methods for the given backend.

    The `timeseries.io.gwf` backend module must define the following methods:

    - `read` : which receives one of more frame files which can be assumed
               to be contiguous, and should return a `TimeSeriesDict`
    - `write` : which receives an output frame file path and a `TimeSeriesDict`
                and does all of the work

    Parameters
    ----------
    backend : `str`
        The name of the backend to register.
        Also accepts `None` to register the ``"gwf"`` format without a
        specific backend; this is applied automatically when this module is
        loaded.
    """
    if backend:
        if backend not in BACKENDS:  # allow self-registration
            BACKENDS.append(backend)
        fmt = f"gwf.{backend.lower()}"
    else:
        fmt = "gwf"

    for klass, reader, writer in (
        (TimeSeries, read_timeseries, write_timeseries),
        (TimeSeriesDict, read_timeseriesdict, write_timeseriesdict),
        (StateVector, read_statevector, write_timeseries),
        (StateVectorDict, read_statevectordict, write_timeseriesdict),
    ):
        if backend:
            message = (
                f"`format='{fmt}` is deprecated, "
                f"please instead use `format='gwf', backend='{backend}'"
            )
            read_ = deprecated_function(
                partial(reader, backend=backend),
                message=message,
            )
            write_ = deprecated_function(
                partial(writer, backend=backend),
                message=message,
            )
        else:
            read_ = reader
            write_ = writer

        klass.read.registry.register_reader(fmt, klass, read_)
        klass.write.registry.register_writer(fmt, klass, write_)
        if backend is None:
            klass.read.registry.register_identifier(
                fmt,
                klass,
                io_gwf.identify_gwf,
            )


# register for timeseries objects
register_gwf_backend(None)  # format='gwf'



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of pelican interaction with `TimeSeries`."""

from unittest import mock

import pytest

from ...testing.errors import pytest_skip_network_error
from .. import TimeSeries

# O3b data for LIGO-Hanford
PELICAN_URL = (
    "osdf:///gwdata/O3b/strain.4k/hdf.v1/H1/1268776960/"
    "H-H1_GWOSC_O3b_4KHZ_R1-1269358592-4096.hdf5",
)
READ_KWARGS = {
    "cache": False,
    "format": "hdf5.gwosc",
}


@pytest_skip_network_error
@pytest.mark.requires("requests_pelican")
def test_timeseries_read_pelican(tmp_path):
    """Check that `TimeSeries.read` can handle Pelican URLs."""
    # force astropy to use tmp_path as the temporary download directory
    with mock.patch("tempfile.gettempdir", return_value=str(tmp_path)):
        data = TimeSeries.read(
            PELICAN_URL,
            **READ_KWARGS,
        )
    assert data.mean().value == pytest.approx(2.768754379315686e-26, abs=1e-28)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GPS time conversion utilities.

The :class:`~astropy.time.Time` object from the astropy package
is imported for user convenience, and a GPS time conversion function
is provided.

All other time conversions can be easily completed using the
:class:`~astropy.time.Time` object.
"""

from typing import TYPE_CHECKING

from astropy.time import Time

from ._ligotimegps import (
    LIGOTimeGPS,
    LIGOTimeGPSLike,
)
from ._tconvert import (
    from_gps,
    tconvert,
    to_gps,
)

if TYPE_CHECKING:
    from ._tconvert import SupportsToGps

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "LIGOTimeGPS",
    "LIGOTimeGPSLike",
    "from_gps",
    "tconvert",
    "to_gps",
]



# Copyright (c) 2013-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.signal.spectral._utils`."""

import pytest
from astropy.units import Unit

from ..spectral import _utils as fft_utils


@pytest.mark.parametrize(("unit", "kwargs", "result"), [
    (Unit("m"), {}, Unit("m^2/Hz")),
    (Unit("m"), {"scaling": "density"}, Unit("m^2/Hz")),
    (Unit("m"), {"scaling": "spectrum"}, Unit("m^2")),
    (None, {}, Unit("Hz^-1")),
])
def test_scale_timeseries_unit(unit, kwargs, result):
    """Test :func:`gwpy.signal.spectral._utils.scale_timeseries_units`."""
    assert fft_utils.scale_timeseries_unit(unit, **kwargs) == result


def test_scale_timeseries_unit_error():
    with pytest.raises(
        ValueError,
        match=r"^unknown scaling: 'other'$",
    ):
        fft_utils.scale_timeseries_unit(None, scaling="other")



# Copyright (c) 2021 Evan Goetz
#               2021-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.transferfunction`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from astropy.time import Time
from numpy.random import default_rng

from ...frequencyseries import FrequencySeries
from .. import TransferFunctionProduct
from .base import _TestFrequencyDomainProduct

if TYPE_CHECKING:
    from typing import ClassVar

__author__ = "Evan Goetz <evan.goetz@ligo.org>"


class TestTransferFunctionProduct(_TestFrequencyDomainProduct):
    """Tests for `gwpy.cli.transferfunction.TransferFunctionProduct`."""

    TEST_CLASS = TransferFunctionProduct
    ACTION = "transferfunction"
    TEST_ARGS: ClassVar[list[str]] = [
        *_TestFrequencyDomainProduct.TEST_ARGS,
        "--chan",
        "Y1:TEST-CHANNEL",
        "--secpfft",
        "0.25",
    ]

    @pytest.fixture
    @classmethod
    def dataprod(cls, prod):
        """Return a `TestTransferFunctionProduct` with data."""
        cls._prod_add_data(prod)
        fftlength = prod.args.secpfft
        for i, ts in enumerate(prod.timeseries):
            nsamp = int(fftlength * 512 / 2.) + 1
            if i % 2 == 0:
                rng = default_rng(i)
                name = f"{prod.timeseries[i+1].name}---{ts.name}"
                fs = FrequencySeries(
                    rng.random(nsamp) + 1j*rng.random(nsamp),
                    x0=0,
                    dx=1/fftlength,
                    channel=prod.timeseries[i+1].channel,
                    name=f"{name}",
                    dtype=complex,
                )
                prod.test_chan = prod.timeseries[i+1].name
            prod.tfs.append(fs)
        return prod

    def test_init(self, prod):
        """Test initialization of `TransferFunctionProduct`."""
        assert prod.chan_list == ["X1:TEST-CHANNEL", "Y1:TEST-CHANNEL"]
        assert prod.ref_chan == prod.chan_list[0]
        assert prod.test_chan == prod.chan_list[1]

    def test_get_suptitle(self, prod):
        """Test suptitle generation."""
        assert prod.get_suptitle() == (
            f"Transfer function: {prod.chan_list[1]}/{prod.chan_list[0]}"
        )

    def test_get_title(self, prod):
        """Test title generation."""
        epoch = prod.start_list[0]
        utc = Time(epoch, format="gps", scale="utc").iso
        t = ", ".join(
            [
                f"{utc} | {epoch} ({prod.duration})",
                f"fftlength={prod.args.secpfft}",
                f"overlap={prod.args.overlap}",
            ],
        )
        assert prod.get_title() == t

    def test_set_plot_properties(self, plotprod):
        """Test setting plot properties."""
        assert plotprod.ax.get_title() == (plotprod.args.title or "")
        assert plotprod.plot.get_suptitle() == (plotprod.args.suptitle or "")



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test :mod:`gwpy.timeseries.io.gwf.utils`."""

import pytest

from ..io.gwf import utils as io_gwf_utils

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize(("value", "channels", "expected_type", "result"), [
    pytest.param(
        10,
        ["A", "B"],
        None,
        {"A": 10, "B": 10},
        id="single",
    ),
    pytest.param(
        (10, 20),
        ["A", "B"],
        None,
        {"A": 10, "B": 20},
        id="zip",
    ),
    pytest.param(
        {"A": 10, "B": 20, "C": 30},
        ["A", "B"],
        None,
        {"A": 10, "B": 20},
        id="dict",
    ),
    pytest.param(
        [10, 20],
        ["A", "B"],
        list,
        {"A": [10, 20], "B": [10, 20]},
        id="iterable",
    ),
    pytest.param(
        [[10, 20], [30, 40]],
        ["A", "B"],
        list,
        {"A": [10, 20], "B": [30, 40]},
        id="nested",
    ),
])
def test_channel_dict_kwarg(value, channels, expected_type, result):
    """Test `_channel_dict_kwarg()`."""
    assert io_gwf_utils._channel_dict_kwarg(
        value,
        channels,
        expected_type=expected_type,
    ) == result


def test_channel_dict_kwarg_error():
    """Test `_channel_dict_kwarg()` zip error handling."""
    with pytest.raises(
        ValueError,
        match="TEST is shorter than channels list",
    ):
        io_gwf_utils._channel_dict_kwarg(
            [1, 2, 3],
            ["A", "B", "C", "D"],
            varname="TEST",
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Channel` and `ChannelList`."""

from __future__ import annotations

import re
from copy import copy
from math import ceil
from typing import (
    TYPE_CHECKING,
    overload,
)

import numpy
from astropy import units
from urllib3.util import parse_url

from ..io import nds2 as io_nds2
from ..io.registry import UnifiedReadWriteMethod
from ..time import to_gps
from .connect import (
    ChannelListRead,
    ChannelListWrite,
)
from .units import parse_unit

if TYPE_CHECKING:
    import builtins
    from typing import (
        Literal,
        Self,
        SupportsFloat,
    )

    import arrakis
    import nds2
    from astropy.units import (
        Quantity,
        UnitBase,
    )

    from ..segments import SegmentListDict
    from ..time import SupportsToGps
    from ..typing import UnitLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# NOTE: when we depend on numpy>=2.0, numpy.bool_ here can
#       be replaced with numpy.bool.
BOOL_TYPES = {bool, numpy.bool_, numpy.dtype(bool), "bool"}

QUOTE_REGEX = re.compile(r'^[\s\"\']+|[\s\"\']+$')


class Channel:
    """Representation of a gravitational-wave detectory data channel.

    Parameters
    ----------
    name : `str`, `Channel`
        Name of this Channel (or another  Channel itself).
        If a `Channel` is given, all other parameters not set explicitly
        will be copied over.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        Rate at which data are sampled for this channel, simple floats
        must be given in Hz.

    unit : `~astropy.units.Unit`, `str`, optional
        Name of the unit for the data of this channel.

    frequency_range : `tuple` of `float`
        ``[low, high)`` spectral frequency range of interest for this channel.

    safe : `bool`, optional
        If `True` this channel is 'safe' to use as a witness of
        non-gravitational-wave noise in the detector.

    dtype : `numpy.dtype`, optional
        Numeric type of data for this channel.

    frametype : `str`, optional
        LDAS name for frames that contain this channel.

    model : `str`, optional
        Name of the SIMULINK front-end model that produces this channel.

    Notes
    -----
    The `Channel` structure implemented here is designed to match the
    data recorded in the LIGO Channel Information System
    (https://cis.ligo.org) for which a query interface is provided.
    """

    MATCH = re.compile(
        r"((?:(?P<ifo>[A-Z]\d))?|[\w-]+):"  # match IFO prefix
        r"(?:(?P<system>[a-zA-Z0-9]+))?"  # match system
        r"(?:[-_](?P<subsystem>[a-zA-Z0-9]+))?"  # match subsystem
        r"(?:[-_](?P<signal>[a-zA-Z0-9_-]+?))?"  # match signal
        r"(?:[\.-](?P<trend>[a-z]+))?"  # match trend type
        r"(?:,(?P<type>([a-z]-)?[a-z]+))?$",  # match channel type
    )

    def __init__(
        self,
        name: str | Self,
        **params,
    ) -> None:
        """Create a new `Channel`.

        Parameters
        ----------
        name : `str`, `Channel`
            The name of the new channel, or an existing channel to copy.

        params
            ``(key, value)`` pairs for attributes of new channel.
        """
        # init properties
        self._name: str  # set in _init_name()
        self._ifo: str | None = None
        self._system: str | None = None
        self._subsystem: str | None = None
        self._signal: str | None = None
        self._trend: str | None = None
        self._sample_rate: Quantity | None = None
        self._unit: UnitBase | None = None
        self._frequency_range: Quantity | None = None
        self._safe: bool | None = None
        self._type: str | None = None
        self._dtype: numpy.dtype | None = None
        self._frametype: str | None = None
        self._model: str | None = None
        self._url: str | None = None

        # copy existing Channel
        if isinstance(name, Channel):
            self._init_from_channel(name)

        # parse name into component parts
        else:
            self._init_name(name)

        # set metadata
        for key, value in params.items():
            setattr(self, key, value)

    def _init_from_channel(self, other: Self) -> None:
        """Copy attributes from ``other`` to this channel."""
        for key, value in vars(other).items():
            setattr(self, key, copy(value))

    def _init_name(self, name: str) -> None:
        """Initialise the name of this `Channel`."""
        try:
            parts = self.parse_channel_name(name)
        except (TypeError, ValueError):  # failed to parse
            self.name = str(name)
        else:
            self.name = str(name).split(",")[0]
            for key, val in parts.items():
                try:
                    setattr(self, key, val)
                except AttributeError:
                    setattr(self, f"_{key}", val)

    # -- properties -----------------------------

    @property
    def name(self) -> str:
        """Name of this channel.

        This should follow the naming convention, with the following
        format: 'IFO:SYSTEM-SUBSYSTEM_SIGNAL'
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = str(name)

    @property
    def sample_rate(self) -> Quantity | None:
        """Rate of samples (Hertz) for this channel."""
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, rate: SupportsFloat | Quantity | None) -> None:
        if rate is None:
            self._sample_rate = None
        elif isinstance(rate, units.Quantity):
            self._sample_rate = rate
        else:
            # pylint: disable=no-member
            self._sample_rate = units.Quantity(float(rate), unit=units.Hertz)

    @sample_rate.deleter
    def sample_rate(self) -> None:
        self._sample_rate = None

    @property
    def unit(self) -> UnitBase | None:
        """Data unit for this channel."""
        return self._unit

    @unit.setter
    def unit(self, u: UnitLike) -> None:
        if u is None:
            self._unit = None
        else:
            self._unit = parse_unit(u)

    @unit.deleter
    def unit(self) -> None:
        self._unit = None

    @property
    def frequency_range(self) -> Quantity | None:
        """Frequency range of interest (Hertz) for this channel."""
        return self._frequency_range

    @frequency_range.setter
    def frequency_range(self, frange: tuple[float, float] | None) -> None:
        if frange is None:
            self._frequency_range = None
        else:
            low, hi = frange
            self._frequency_range = units.Quantity((low, hi), unit="Hz")

    @frequency_range.deleter
    def frequency_range(self) -> None:
        self._frequency_range = None

    @property
    def safe(self) -> bool | None:
        """Whether this channel is 'safe' to use as a noise witness.

        Any channel that records part or all of a GW signal as it
        interacts with the interferometer is not safe to use as a noise
        witness

        A safe value of `None` simply indicates that the safety of this
        channel has not been determined
        """
        return self._safe

    @safe.setter
    def safe(self, s: bool | None) -> None:
        if s is None:
            self._safe = None
        else:
            self._safe = bool(s)

    @safe.deleter
    def safe(self) -> None:
        self._safe = None

    @property
    def model(self) -> str | None:
        """Name of the SIMULINK front-end model that defines this channel."""
        return self._model

    @model.setter
    def model(self, mdl: str | None) -> None:
        if mdl is None:
            self._model = None
        else:
            self._model = mdl.lower()

    @model.deleter
    def model(self) -> None:
        self._model = None

    @property
    def type(self) -> str | None:
        """DAQ data type for this channel.

        Valid values for this field are restricted to those understood
        by the NDS2 client sofware, namely:

        'm-trend', 'online', 'raw', 'reduced', 's-trend', 'static', 'test-pt'
        """
        return self._type

    @type.setter
    def type(self, type_: str | int | None) -> None:
        if type_ is None:
            self._type = None
        else:
            self._type = io_nds2.Nds2ChannelType.find(type_).nds2name

    @type.deleter
    def type(self) -> None:
        self._type = None

    @property
    def ndstype(self) -> int:
        """NDS type integer for this channel.

        This property is mapped to the `Channel.type` string.
        """
        if self.type is None:
            return 0  # UNKNOWN
        return io_nds2.Nds2ChannelType.find(self.type).value

    @ndstype.setter
    def ndstype(self, type_: str | int | None) -> None:
        self.type = type_

    @ndstype.deleter
    def ndstype(self) -> None:
        del self.type

    @property
    def dtype(self) -> numpy.dtype | None:
        """Numeric type for data in this channel."""
        return self._dtype

    @dtype.setter
    def dtype(self, type_: str | int | builtins.type | None) -> None:
        if type_ is None:
            self._dtype = None
        elif type_ in BOOL_TYPES:  # NDS2 doesn't support bool
            self._dtype = numpy.dtype(bool)
        else:
            self._dtype = io_nds2.Nds2DataType.find(type_).dtype

    @dtype.deleter
    def dtype(self) -> None:
        self._dtype = None

    @property
    def url(self) -> str | None:
        """CIS browser url for this channel."""
        return self._url

    @url.setter
    def url(self, href: str | None) -> None:
        if href is None:
            self._url = None
        else:
            url = parse_url(href)
            if url.scheme not in ("http", "https", "file"):
                msg = f"Invalid URL '{href}'"
                raise ValueError(msg)
            self._url = href

    @url.deleter
    def url(self) -> None:
        self._url = None

    @property
    def frametype(self) -> str | None:
        """LDAS type description for frame files containing this channel."""
        return self._frametype

    @frametype.setter
    def frametype(self, ft: str | None) -> None:
        self._frametype = ft

    @frametype.deleter
    def frametype(self) -> None:
        self._frametype = None

    # -- read-only properties -------------------

    @property
    def ifo(self) -> str | None:
        """Interferometer prefix for this channel."""
        return self._ifo

    @property
    def system(self) -> str | None:
        """Instrumental system for this channel."""
        return self._system

    @property
    def subsystem(self) -> str | None:
        """Instrumental sub-system for this channel."""
        return self._subsystem

    @property
    def signal(self) -> str | None:
        """Instrumental signal for this channel."""
        return self._signal

    @property
    def trend(self) -> str | None:
        """Trend type for this channel."""
        return self._trend

    @property
    def texname(self) -> str:
        """Name of this channel in LaTeX printable format."""
        return str(self).replace("_", r"\_")

    @property
    def ndsname(self) -> str:
        """Name of this channel as stored in the NDS database."""
        if self.type not in {None, "raw", "reduced", "online"}:
            return f"{self.name},{self.type}"
        return self.name

    # -- classmethods ---------------------------

    @classmethod
    def query(
        cls,
        name: str,
        *,
        kerberos: bool | None = None,
        **kwargs,
    ) -> Self:
        """Query the LIGO Channel Information System for ``name``.

        Parameters
        ----------
        name : `str`
            Name of channel for which to query.

        kerberos : `bool`, optional
            If `True` use an existing Kerberos ticket as the authentication credential.
            Default (`None`) is to check for credentials and request username
            and password if none are found.

        kwargs
            Other keyword arguments are passed directly to
            :func:`ciecplib.get`.

        Returns
        -------
        channel : `Channel`
            A new `Channel` containing all of the attributes set from
            its entry in the CIS.
        """
        channellist = ChannelList.query(name, kerberos=kerberos, **kwargs)
        if not channellist:
            msg = f"No channels found matching '{name}'"
            raise ValueError(msg)
        if len(channellist) > 1:
            msg = (
                f"{len(channellist)} channels found matching '{name}', "
                "please refine search, or use `ChannelList.query` to "
                "return all results"
            )
            raise ValueError(msg)
        return channellist[0]

    @classmethod
    def query_nds2(
        cls,
        name: str,
        host: str | None = None,
        port: int | None = None,
        connection: nds2.connection | None = None,
        type: str | int | None = None,  # noqa: A002
    ) -> Self:
        """Query an NDS server for channel information.

        Parameters
        ----------
        name : `str`
            Name of channel for which to query.

        host : `str`, optional
            Name of NDS2 server.

        port : `int`, optional
            Port number for NDS2 connection.

        connection : `nds2.connection`
            Open connection to use for query.

        type : `str`, `int`
            NDS2 channel type with which to restrict query

        Returns
        -------
        channel : `Channel`
            Channel with metadata retrieved from NDS2 server.

        Raises
        ------
        ValueError
            If multiple channels are found for a given name.

        Notes
        -----
        .. warning::

            One of ``host`` or ``connection`` is required.
        """
        return ChannelList.query_nds2(
            [name],
            host=host,
            port=port,
            connection=connection,
            type=type,
            unique=True,
        )[0]

    @classmethod
    def from_nds2(
        cls,
        nds2channel: nds2.channel,
    ) -> Self:
        """Generate a new channel using an existing nds2.channel object."""
        return cls(
            nds2channel.name,
            sample_rate=nds2channel.sample_rate,
            unit=nds2channel.signal_units or None,
            dtype=nds2channel.data_type,
            type=nds2channel.channel_type,
        )

    @classmethod
    def from_arrakis(
        cls,
        arrakischannel: arrakis.Channel,
    ) -> Self:
        """Generate a new channel using an existing `arrakis.Channel`.

        Parameters
        ----------
        arrakischannel : `arrakis.Channel`
            The input channel from Arrakis to parse.

        Returns
        -------
        channel : `gwpy.detector.Channel`
            A new `Channel`.
        """
        return cls(
            arrakischannel.name,
            sample_rate=arrakischannel.sample_rate,
            dtype=arrakischannel.data_type,
        )

    # -- methods --------------------------------

    @classmethod
    def parse_channel_name(
        cls,
        name: str,
        *,
        strict: bool = True,
    ) -> dict[str, str | None]:
        """Decompose a channel name string into its components.

        This method parses channels acccording to the LIGO Channel Naming
        Convention :dcc:`LIGO-T990033`.

        Parameters
        ----------
        name : `str`
            Name to parse.

        strict : `bool`, optional
            If `True` (default) require exact matching of format,
            with no surrounding text.

        Returns
        -------
        match : `dict`
            `dict` of channel name components with the following keys:

            - `'ifo'`: the letter-number interferometer prefix
            - `'system'`: the top-level system name
            - `'subsystem'`: the second-level sub-system name
            - `'signal'`: the remaining underscore-delimited signal name
            - `'trend'`: the trend type
            - `'ndstype'`: the NDS2 channel suffix

            Any optional keys that aren't found will return a value of `None`

        Raises
        ------
        ValueError
            if the name cannot be parsed with at least an IFO and SYSTEM

        Examples
        --------
        >>> Channel.parse_channel_name("L1:LSC-DARM_IN1_DQ")
        {'ifo': 'L1',
         'ndstype': None,
         'signal': 'IN1_DQ',
         'subsystem': 'DARM',
         'system': 'LSC',
         'trend': None}

        >>> Channel.parse_channel_name(
        ...     "H1:ISI-BS_ST1_SENSCOR_GND_STS_X_BLRMS_100M_300M.rms,m-trend",
        ... )
        {'ifo': 'H1',
         'ndstype': 'm-trend',
         'signal': 'ST1_SENSCOR_GND_STS_X_BLRMS_100M_300M',
         'subsystem': 'BS',
         'system': 'ISI',
         'trend': 'rms'}
        """
        match = cls.MATCH.search(name)
        if match is None or (
            strict
            and (match.start() != 0 or match.end() != len(name))
        ):
            msg = "Cannot parse channel name according to LIGO-T990033"
            raise ValueError(msg)
        return match.groupdict()

    @overload
    def find_frametype(
        self,
        gpstime: SupportsToGps | None = None,
        *,
        frametype_match: str | re.Pattern | None = None,
        return_all: Literal[False] = False,
        allow_tape: bool = True,
        **kwargs,
    ) -> str: ...

    @overload
    def find_frametype(
        self,
        gpstime: SupportsToGps | None = None,
        *,
        frametype_match: str | re.Pattern | None = None,
        return_all: Literal[True] = True,
        allow_tape: bool = True,
        **kwargs,
    ) -> list[str]: ...

    def find_frametype(
        self,
        gpstime: SupportsToGps | None = None,
        *,
        frametype_match: str | re.Pattern | None = None,
        return_all: bool = False,
        allow_tape: bool = True,
        **kwargs,
    ) -> str | list[str]:
        """Find the containing frametype(s) for this `Channel`.

        Parameters
        ----------
        gpstime : `int`, optional
            A reference GPS time at which to search for frame files.
            Default is to search in the latest available data for each
            discoverable dataset.

        frametype_match : `str`
            A regular expression string to use to down-select from the
            list of all available datasets.

        return_all: `bool`, optional
            If `True` return all matched datasets; if `False` (default)
            only the first match is returned

        allow_tape : `bool`, default: `True`
            If `True` (default) include datasets whose files are stored on slow
            magnetic tape.

        kwargs
            Other keyword arguments are passed directly to
            :func:`gwpy.io.datafind.find_frametype`.

        Returns
        -------
        frametype : `str` or `list[str]`
            If ``return_all=False`` a single `str` containing the 'best' dataset name.
            If ``return_all=True`` a `list` of dataset names.
        """
        from ..io import datafind

        return datafind.find_frametype(
            self,
            gpstime=gpstime,
            frametype_match=frametype_match,
            return_all=return_all,
            allow_tape=allow_tape,
            **kwargs,
        )

    def copy(self) -> Self:
        """Return a copy of this `Channel`."""
        return type(self)(self)

    def __str__(self) -> str:
        """Return the name of this `Channel`."""
        return self.name or ""

    def __repr__(self) -> str:
        """Return a printable representation of this `Channel`."""
        repr_ = f'<Channel("{self}"'
        if self.type:
            repr_ += f" [{self.type}]"
        repr_ += f", {self.sample_rate}"
        return repr_ + f") at {hex(id(self))}>"

    def __eq__(self, other: object) -> bool:
        """Return `True` if all attributes of this channel match ``other``."""
        try:
            for attr in ("name", "sample_rate", "unit", "url", "type", "dtype"):
                if getattr(self, attr) != getattr(other, attr):
                    return False
        except (
            AttributeError,  # no such attribute
            TypeError,  # attribute values can't be compared
        ):
            return False
        return True

    def __hash__(self) -> int:
        """Return a hash of this `Channel`."""
        hash_ = 0
        for attr in ("name", "sample_rate", "unit", "url", "type", "dtype"):
            hash_ += hash(getattr(self, attr))
        return hash_


class ChannelList(list):
    """A `list` of `channels <Channel>`, with parsing utilities."""

    # -- properties ------------------

    @property
    def ifos(self) -> set[str]:
        """The `set` of interferometer prefixes used in this `ChannelList`."""
        return {c.ifo for c in self}

    # -- i/o -------------------------

    read = UnifiedReadWriteMethod(ChannelListRead)
    write = UnifiedReadWriteMethod(ChannelListWrite)

    # -- methods ---------------------

    @classmethod
    def from_names(cls, *names: str) -> Self:
        """Create a new `ChannelList` from a list of names.

        The list of names can include comma-separated sets of names,
        in which case the return will be a flattened list of all parsed
        channel names.
        """
        new = cls()
        for namestr in names:
            for name in cls._split_names(namestr):
                new.append(Channel(name))
        return new

    @staticmethod
    def _split_names(namestr: str) -> list[str]:
        """Split a comma-separated list of channel names."""
        out = []
        namestr = QUOTE_REGEX.sub("", namestr)
        while True:
            namestr = namestr.strip("' \n")
            if "," not in namestr:
                break
            for nds2type in [*io_nds2.Nds2ChannelType.nds2names(), ""]:
                if nds2type and f",{nds2type}" in namestr:
                    try:
                        channel, ctype, namestr = namestr.split(",", 2)
                    except ValueError:
                        channel, ctype = namestr.split(",")
                        namestr = ""
                    out.append(f"{channel},{ctype}")
                    break
                if nds2type == "" and "," in namestr:
                    channel, namestr = namestr.split(",", 1)
                    out.append(channel)
                    break
        if namestr:
            out.append(namestr)
        return out

    def find(self, name: str) -> int:
        """Find the `Channel` with a specific name in this `ChannelList`.

        Parameters
        ----------
        name : `str`
            name of the `Channel` to find

        Returns
        -------
        index : `int`
            the position of the first `Channel` in this `ChannelList`
            whose `~Channel.name` matches the search key.

        Raises
        ------
        ValueError
            if no matching `Channel` is found.
        """
        for i, chan in enumerate(self):
            if name == chan.name:
                return i
        raise ValueError(name)

    def sieve(
        self,
        name: str | re.Pattern | None = None,
        sample_rate: float | None = None,
        *,
        sample_range: tuple[float, float] | None = None,
        exact_match: bool = False,
        **others,
    ) -> Self:
        """Find all channels in this list matching the specified criteria.

        Parameters
        ----------
        name : `str`, or regular expression
            Any part of the channel name against which to match
            (or full name if `exact_match=False` is given).

        sample_rate : `float`
            Rate (number of samples per second) to match exactly.

        sample_range : 2-`tuple`
            `[low, high]` closed interval or rates to match within.

        exact_match : `bool`
            If `True` return channels matching ``name`` exactly.
            If `False` (default) allow partial matches.

        others:
            Other ``(key, value)`` attribute pairs to match.

        Returns
        -------
        new : `ChannelList`
            A new `ChannelList` containing the matching channels.
        """
        # format name regex
        if isinstance(name, re.Pattern):
            flags = name.flags
            name = str(name.pattern)
        else:
            flags = 0
        if name is not None:
            if exact_match:
                name = name if name.startswith(r"\A") else fr"\A{name}"
                name = name if name.endswith(r"\Z") else fr"{name}\Z"
            name_regexp = re.compile(name, flags=flags)

        def _match(channel: Channel) -> bool:
            if name is not None and name_regexp.search(channel.name) is None:
                return False
            if sample_rate is not None and channel.sample_rate != sample_rate:
                return False
            if sample_range is not None and (
                channel.sample_rate is None
                or sample_range[0] > channel.sample_rate.value
                or sample_range[1] <= channel.sample_rate.value
            ):
                return False
            for key, val in others.items():
                if val is not None and getattr(channel, key, None) != val:
                    return False
            return True

        return type(self)(filter(_match, self))

    @classmethod
    def query(
        cls,
        name: str,
        *,
        kerberos: bool | None = None,
        **kwargs,
    ) -> Self:
        """Query the LIGO Channel Information System a `ChannelList`.

        Parameters
        ----------
        name : `str`
            name of channel, or part of it.

        kerberos : `bool`, optional
            If `True` use an existing Kerberos ticket as the authentication credential.
            Default (`None`) is to check for credentials and request username
            and password if none are found.

        kwargs
            Other keyword arguments are passed directly to
            :func:`ciecplib.get`.

        Returns
        -------
        channels : `ChannelList`
            A new list containing all `Channels <Channel>` found.
        """
        from .io import cis
        return cls(cis.query(
            name,
            kerberos=kerberos,
            **kwargs,
        ))

    @classmethod
    def query_nds2(
        cls,
        names: list[str],
        *,
        host: str | None = None,
        port: int | None = None,
        connection: nds2.connection | None = None,
        type: str | int | None = io_nds2.NDS2_CHANNEL_TYPE_ANY,  # noqa: A002
        unique: bool = False,
    ) -> Self:
        """Query an NDS server for channel information.

        Parameters
        ----------
        names : `str`
            Names of requested channels.

        host : `str`, optional
            Name of NDS2 server.

        port : `int`, optional
            Port number for NDS2 connection.

        connection : `nds2.connection`, optional
            Open connection to use for query.

        type : `str`, `int`, optional
            NDS2 channel type with which to restrict query

        unique : `bool`, optional
            If `True` require a unique query result for each name given.
            Default is `False`.

        Returns
        -------
        channellist : `ChannelList`
            list of `Channels <Channel>` with metadata retrieved from
            NDS2 server

        Raises
        ------
        ValueError
            If multiple channels are found for a given name and `unique=True`
            is given.

        Notes
        -----
        .. warning::

            One of ``host`` or ``connection`` is required.
        """
        ndschannels = io_nds2.find_channels(
            names,
            host=host,
            port=port,
            connection=connection,
            type=type,
            unique=unique,
        )
        return cls(map(Channel.from_nds2, ndschannels))

    @classmethod
    def query_nds2_availability(
        cls,
        channels: list[str | Channel],
        start: SupportsToGps,
        end: SupportsToGps,
        ctype: int | str = io_nds2.Nds2ChannelType.any().value,
        connection: nds2.connection | None = None,
        host: str | None = None,
        port: int | None = None,
    ) -> SegmentListDict:
        """Query for when data are available for these channels in NDS2.

        Parameters
        ----------
        channels : `list` of `str` or `Channel`
            List of `Channel` or `str` for which to search.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of search.
            Any input parseable by `~gwpy.time.to_gps` is fine.

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of search.
            Any input parseable by `~gwpy.time.to_gps` is fine.

        ctype : `int`, `str`
            The NDS2 channel type name or enum ID against which to restrict
            results. Default is ``127`` which means 'all'.

        host : `str`, optional
            Name of NDS2 server to use.

        port : `int`, optional
            Port number for NDS2 connection.

        connection : `nds2.connection`, optional
            Open connection to use for query.

        Returns
        -------
        segdict : `~gwpy.segments.SegmentListDict`
            dict of ``(name, SegmentList)`` pairs
        """
        start = int(to_gps(start))
        end = ceil(to_gps(end))
        with io_nds2._connection(connection=connection, host=host, port=port) as conn:
            chans = io_nds2.find_channels(
                channels,
                unique=True,
                epoch=(start, end),
                type=ctype,
                connection=conn,
            )
            availability = io_nds2.get_availability(
                chans,
                start,
                end,
                connection=conn,
            )
        return type(availability)(zip(
            channels,
            availability.values(),
            strict=True,
        ))



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for command-line tools for GWpy."""

from __future__ import annotations

import argparse
import inspect
import logging
import os
import textwrap
from argparse import (
    ArgumentDefaultsHelpFormatter,
    RawDescriptionHelpFormatter,
)
from typing import TYPE_CHECKING

from ..log import init_logger

if TYPE_CHECKING:
    from argparse import (
        Action,
        _MutuallyExclusiveGroup,
    )
    from collections.abc import (
        Iterable,
        Mapping,
    )
    from logging import Logger
    from typing import TypeVar

    _ActionT = TypeVar("_ActionT", bound=Action)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def is_manpage() -> bool:
    """Return `True` if being called to generate a man page."""
    return any(
        "argparse_manpage" in frame.filename
        for frame in inspect.stack()
    )


def get_logger(name: str, fallback: str = "__main__") -> logging.Logger:
    """Get a logger for the given name."""
    if name == "__main__":
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])
        spec = getattr(mod, "__spec__", None)
        if spec is not None:
            name = spec.name
        else:
            name = fallback
    return logging.getLogger(name)


def init_verbose_logging(
    name: str = "gwpy",
    verbosity: int = 0,
) -> Logger:
    """Configure logging based on a verbosity count."""
    # If user did not specify verbosity, don't change anything;
    # this allows the logging level to be configured by other means,
    # e.g. a config file or environment variable.
    if not verbosity:
        return init_logger(name)

    # Otherwise, set the level for the gwpy logger based on verbosity
    level = max(3 - verbosity, 0) * 10
    return init_logger(name, level=level)


def parse_option(value: str) -> tuple[str, bool | float | str | None]:
    """Parse a key=value option string into a key and value tuple.

    This function attempts to convert the value to a boolean or numeric
    type if possible; otherwise, it is returned as a string.

    Parameters
    ----------
    value : str
        The option string in the format "key=value".

    Returns
    -------
    key, value: tuple[str, bool | float | str | None]
        The parsed key and value.

    Examples
    --------
    >>> parse_option("threshold=0.5")
    ('threshold', 0.5)
    >>> parse_option("enable_feature=true")
    ('enable_feature', True)
    >>> parse_option("name=example")
    ('name', 'example')
    """
    key, val = value.split("=", 1)

    # Handle boolean and None
    if (vall := val.lower()) in ("true", "yes"):
        return key, True
    if vall in ("false", "no"):
        return key, False
    if vall == "none":
        return key, None

    # Handle numeric
    try:
        val = float(val)
    except ValueError:
        # String
        return key, val
    if val.is_integer():
        val = int(val)
    return key, val


def parse_options_dict(options: list[str]) -> dict[str, bool | float | str | None]:
    """Parse a list of key=value option strings into a dictionary.

    Parameters
    ----------
    options : list[str]
        A list of option strings in the format "key=value".

    Returns
    -------
    dict[str, str | bool | int | float]
        A dictionary of parsed key-value pairs.

    Examples
    --------
    >>> parse_options_dict(["threshold=0.5", "enable_feature=true", "name=example"])
    {'threshold': 0.5, 'enable_feature': True, 'name': 'example'}
    """
    out: dict[str, bool | float | str | None ] = {}
    for optstr in options:
        key, val = parse_option(optstr)
        out[key] = val
    return out


class HelpFormatter(
    ArgumentDefaultsHelpFormatter,
    RawDescriptionHelpFormatter,
):
    """Custom help formatter for GWpy tools."""

    def _format_usage(
        self,
        usage: str | None,
        actions: Iterable[Action],
        groups: Iterable[_MutuallyExclusiveGroup],
        prefix: str | None,
    ) -> str:
        """Format the usage string for the help message."""
        if prefix is None:
            prefix = "Usage: "
        return super()._format_usage(
            usage,
            actions,
            groups,
            prefix,
        )


class ArgumentParser(argparse.ArgumentParser):
    """Custom argument parser for GWpy tools."""

    def __init__(self, **kwargs) -> None:
        """Initialize the argument parser."""
        # Extract docstring parameters
        description = kwargs.pop("description", None)
        epilog = kwargs.pop("epilog", None)
        examples: Mapping[str, str | list[str]] | None = kwargs.pop("examples", {})
        manpage: list[dict[str, str]] = kwargs.pop("manpage", [])

        # Format the docstring
        if description is not None:
            description, epilog, manpage_sections = (
                self._compile_docstring(
                    description,
                    epilog,
                    examples,
                )
            )
            manpage.extend(manpage_sections)

        # Initialize the base class
        kwargs.setdefault("formatter_class", HelpFormatter)
        super().__init__(
            description=description,
            epilog=epilog,
            **kwargs,
        )

        # Rename the argument groups to use title case
        self._positionals.title = "Positional arguments"
        self._optionals.title = "Options"

        # Add manpage generation if requested
        self._manpage = manpage

    def _format_examples(self, examples: Mapping[str, str | list[str]]) -> str:
        """Format examples for the docstring."""
        # Manpage format
        if is_manpage():
            lines: list[str] = []
            for desc, cmd in examples.items():
                if isinstance(cmd, str):
                    cmd = cmd.strip().splitlines()
                lines.extend((
                    r".IP \[bu]",
                    fr"\fB{desc}:\fR",
                    ".sp",
                    r".RS 4",
                    ".nf",
                    *(f"$ {line}" for line in cmd),
                    ".fi",
                    r".RE",
                ))
            return os.linesep.join(lines)
        # Argparse help format
        indnt = "  "
        lines = ["Examples:"]
        for desc, cmd in examples.items():
            lines.extend((
                "",
                textwrap.indent(f"{desc}:", indnt),
                "",
                textwrap.indent(f"$ {cmd}", indnt * 2),
            ))
        return os.linesep.join(lines)

    def _compile_docstring(
        self,
        description: str,
        epilog: str | None,
        examples: Mapping[str, str | list[str]] | None,
    ) -> tuple[str, str | None, list[dict[str, str]]]:
        """Compile the docstring for this tool, either for a manual page, or --help."""
        examples_doc = self._format_examples(examples or {})
        if examples and is_manpage():
            if epilog:
                description += os.linesep * 2 + epilog
            manpage_sections = [
                {
                    "heading": "examples",
                    "content": examples_doc,
                },
            ]
            return description, "", manpage_sections
        if examples and epilog:
            epilog = examples_doc + os.linesep * 2 + epilog
        elif examples:
            epilog = examples_doc
        return description, epilog, []

    def _add_action(self, action: _ActionT) -> _ActionT:
        """Add an action to the parser, combining groups for manpages."""
        if is_manpage():
            return self._optionals._add_action(action)
        return super()._add_action(action)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous utilties for GWpy."""

from __future__ import annotations

import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
    )
    from typing import TypeVar

    T = TypeVar("T")
    R = TypeVar("R")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def if_not_none(
    func: Callable[[T], R],
    value: T,
) -> R | None:
    """Apply func to value if value is not None.

    Examples
    --------
    >>> from gwpy.utils.misc import if_not_none
    >>> if_not_none(int, '1')
    1
    >>> if_not_none(int, None)
    None
    """
    if value is None:
        return None
    return func(value)


def round_to_power(
    x: float,
    base: float = 2,
    which: str | None = None,
) -> float:
    """Round a positive value to the nearest integer power of `base`.

    Parameters
    ----------
    x : `float`
        Value to round, must be strictly positive.

    base : `float`
        Base to whose power `x` will be rounded.

    which : `str` or `None`
        Which value to round to, must be one of `'lower'`, `'upper'`, or
        `None` to round to whichever is nearest.

    Returns
    -------
    rounded : `float`
        The rounded value.

    Notes
    -----
    The object returned will be of the same type as `base`.

    Examples
    --------
    >>> from gwpy.utils.misc import round_to_power
    >>> round_to_power(2)
    2
    >>> round_to_power(9, base=10)
    10
    >>> round_to_power(5, which='lower')
    4
    """
    selector: Callable
    if which == "lower":
        selector = math.floor
    elif which == "upper":
        selector = math.ceil
    elif which is None:
        selector = round
    else:
        msg = "'which' argument must be one of 'lower', 'upper', or None"
        raise ValueError(msg)
    return type(base)(base ** selector(math.log(x, base)))


def property_alias(
    prop: property,
    doc: str | None = None,
) -> property:
    """Create a property alias for another property.

    This is useful when you want to expose a property from a contained
    object as a property of the containing object, but want to be able
    to set a different docstring for the alias.

    Parameters
    ----------
    prop : `property`
        The property to alias.

    doc : `str`, optional
        The docstring for the new property.  If not given, the docstring
        of the original property will be used.

    Returns
    -------
    alias : `property`
        The new property which acts as an alias for `prop`.

    Examples
    --------
    >>> from gwpy.utils.misc import property_alias
    >>> class A:
    ...     @property
    ...     def x(self):
    ...         "The x property"
    ...         return 1
    >>> class B:
    ...     def __init__(self):
    ...         self.a = A()
    ...     a = property_alias(A.x, "Alias for A.x")
    >>> b = B()
    >>> b.a
    1
    >>> B.a.__doc__
    'Alias for A.x'
    """
    return property(
        prop.fget,
        prop.fset,
        prop.fdel,
        doc or prop.__doc__,
    )


def unique(
    list_: Iterable[T],
) -> list[T]:
    """Return a version of ``list_`` with unique elements, preserving order.

    Examples
    --------
    >>> from gwpy.utils.misc import unique
    >>> unique(['b', 'c', 'a', 'a', 'd', 'e', 'd', 'a'])
    ['b', 'c', 'a', 'd', 'e']
    """
    return list(dict.fromkeys(list_))



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test of `Table.read` and `Table.write` with ``format='hdf5'``."""

import pytest

from ...testing.utils import assert_table_equal
from ..filter import filter_table
from .utils import (
    TABLE_CLASSES,
    random_table,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize("table_class", TABLE_CLASSES[1:])
def test_read_write_hdf5(table_class, tmp_path):
    """Test `EventTable.read` and `EventTable.write` with ``format='hdf5'``."""
    table = random_table(
        names=["time", "frequency", "snr"],
        table_class=table_class,
    )

    # check that we can write a table and read it back
    tmp = tmp_path / "table.h5"
    table.write(tmp, path="/data")
    t2 = type(table).read(tmp, path="/data")
    assert_table_equal(t2, table)

    # check that we can read with columns and where
    t2 = type(table).read(
        tmp,
        path="/data",
        where="frequency>500",
        columns=["time", "snr"],
    )
    assert_table_equal(
        t2,
        filter_table(table, "frequency>500")[("time", "snr")],
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.types.array2d`."""

from __future__ import annotations

from typing import (
    Generic,
    TypeVar,
)

import numpy
import pytest
from astropy import units
from numpy.testing import assert_array_equal

from ...segments import Segment
from ...testing import utils
from .. import Array2D, Series
from .test_series import TestSeries as _TestSeries

SEED = 1

Array2DType = TypeVar("Array2DType", bound=Array2D)


class TestArray2D(_TestSeries[Array2D], Generic[Array2DType]):
    """Test `gwpy.types.Array2D`."""

    TEST_CLASS: type[Array2DType] = Array2D

    @classmethod
    def setup_class(cls, dtype=None):
        """Create data with which to test this array."""
        rng = numpy.random.default_rng(seed=SEED)
        cls.data = (rng.random(size=100) * 1e5).astype(
            dtype=dtype,
        ).reshape((20, 5))

    # -- test properties -------------

    def test_y0(self, array: Array2DType):
        """Test `Array2D.y0`."""
        array.y0 = 5
        # test simple
        assert array.y0 == 5 * self.TEST_CLASS._default_yunit

        # test deleter
        del array.y0
        del array.y0
        assert array.y0 == 0 * self.TEST_CLASS._default_yunit

        # test quantity
        array.y0 = 5 * units.m
        assert array.y0 == units.Quantity(5, "m")

    def test_dy(self, array: Array2DType):
        """Test `Array2D.dy`."""
        array.dy = 5 * self.TEST_CLASS._default_yunit
        # test simple
        assert array.dy == units.Quantity(5, self.TEST_CLASS._default_yunit)

        # test deleter
        del array.dy
        del array.dy
        assert array.dy == units.Quantity(1, self.TEST_CLASS._default_yunit)

        # test quantity
        array.dy = 5 * units.m
        assert array.dy == units.Quantity(5, "m")

    def test_yindex(self):
        """Test `Array2D.yindex`."""
        y = numpy.linspace(0, 100, num=self.data.shape[1])

        # test simple
        series = self.create(yindex=y)
        utils.assert_quantity_equal(
            series.yindex,
            units.Quantity(y, self.TEST_CLASS._default_yunit),
        )

        # test deleter
        del series.yindex
        del series.yindex
        y1 = series.y0.value + series.shape[1] * series.dy.value
        y_default = numpy.linspace(
            series.y0.value,
            y1,
            num=series.shape[1],
            endpoint=False,
        )
        utils.assert_quantity_equal(
            series.yindex,
            units.Quantity(y_default, self.TEST_CLASS._default_yunit),
        )

        # test setting of y0 and dy
        series = self.create(yindex=units.Quantity(y, "Farad"))
        assert series.y0 == units.Quantity(y[0], "Farad")
        assert series.dy == units.Quantity(y[1] - y[0], "Farad")
        assert series.yunit == units.Farad
        assert series.yspan == (y[0], y[-1] + y[1] - y[0])

        # test that setting yindex warns about ignoring dy or y0
        with pytest.warns(
            UserWarning,
            match="yindex was given .* dy will be ignored",
        ):
            series = self.create(yindex=units.Quantity(y, "Farad"), dy=1)
        with pytest.warns(
            UserWarning,
            match="yindex was given .* y0 will be ignored",
        ):
            series = self.create(yindex=units.Quantity(y, "Farad"), y0=0)

        # test non-regular yindex
        y = numpy.logspace(0, 2, num=self.data.shape[0])
        series = self.create(yindex=units.Quantity(y, "Mpc"))
        with pytest.raises(AttributeError):
            series.dy  # noqa: B018
        assert series.y0 == units.Quantity(1, "Mpc")
        assert series.yspan == (y[0], y[-1] + y[-1] - y[-2])

    def test_yunit(self):
        """Test `Array2D.yunit`."""
        unit = self.TEST_CLASS._default_yunit
        series = self.create(dy=4 * unit)
        assert series.yunit == unit
        assert series.y0 == 0 * unit
        assert series.dy == 4 * unit
        # for series only, test arbitrary yunit
        if self.TEST_CLASS in (Series, Array2D):
            series = self.create(dy=4, yunit=units.m)
            assert series.y0 == 0 * units.m
            assert series.dy == 4 * units.m

    def test_yspan(self):
        """Test `Array2D.yspan`."""
        # test normal
        series = self.create(y0=1, dy=1)
        assert series.yspan == (1, 1 + 1 * series.shape[1])
        assert isinstance(series.yspan, Segment)
        # test from irregular yindex
        y = numpy.logspace(0, 2, num=self.data.shape[1])
        series = self.create(yindex=y)
        assert series.yspan == (y[0], y[-1] + y[-1] - y[-2])

    def test_transpose(self, array: Array2DType):
        """Test `Array2D.T`."""
        trans = array.T
        utils.assert_array_equal(trans.value, array.value.T)
        assert trans.unit is array.unit
        utils.assert_array_equal(trans.xindex, array.yindex)
        utils.assert_array_equal(trans.yindex, array.xindex)

    # -- test methods ----------------

    @pytest.mark.parametrize("create_kwargs", [
        {"x0": 0, "dx": 1, "y0": 100, "dy": 2},
        {"xindex": numpy.arange(20), "yindex": numpy.linspace(0, 100, 5)},
        {"x0": 0, "dx": 1, "yindex": numpy.linspace(0, 100, 5)},
        {"xindex": numpy.arange(20), "y0": 100, "dy": 2},
    ])
    def test_getitem(self, array: Array2DType, create_kwargs):
        """Test item access and slicing on `Array2D`."""
        array = self.create(name="test_getitem", **create_kwargs)

        # test element returns as quantity
        element = array[0, 0]
        assert element == array[0][0]
        assert isinstance(element, units.Quantity)
        utils.assert_quantity_equal(element, array.value[0, 0] * array.unit)

        # test column slice returns as _columnclass
        utils.assert_quantity_sub_equal(array[2], array[2, :])
        column = array[0, 0::2]
        utils.assert_quantity_sub_equal(
            column,
            self.TEST_CLASS._columnclass(
                array.value[0, 0::2],
                x0=array.y0,
                dx=array.dy * 2,
                name=array.name,
                channel=array.channel,
                unit=array.unit,
                epoch=array.epoch,
            ),
        )

        # test row slice returns as _rowclass
        row = array[1:10:3, 0]
        utils.assert_array_equal(row.value, array.value[1:10:3, 0])
        utils.assert_quantity_sub_equal(
            row,
            self.TEST_CLASS._rowclass(
                array.value[1:10:3, 0],
                x0=array.x0 + array.dx,
                dx=array.dx * 3,
                name=array.name,
                channel=array.channel,
                unit=array.unit,
            ),
            exclude=["epoch"],
        )

        # test dual slice returns type(self) with metadata
        subarray = array[1:5:2, 1:5:2]
        utils.assert_quantity_sub_equal(
            subarray,
            self.TEST_CLASS(
                array.value[1:5:2, 1:5:2],
                x0=array.x0 + array.dx,
                dx=array.dx * 2,
                y0=array.y0 + array.dy,
                dy=array.dy * 2,
                name=array.name,
                channel=array.channel,
                unit=array.unit,
            ),
            exclude=["epoch"],
        )

    def test_single_column_slice(self):
        """Check that we can slice an `Array2D` into a single column.

        But still represent the output as an `Array2D` with `Index` arrays.

        This tests regression of https://gitlab.com/gwpy/gwpy/-/issues/1504.
        """
        # create an array with indices
        a = self.create()
        a.xindex  # noqa: B018
        a.yindex  # noqa: B018

        # select a slice of width 1 (as opposed to indexing a single column)
        b = a[0:1]

        # and check that the index arrays were correctly preserved
        assert isinstance(b, self.TEST_CLASS)
        for attr in ("x0", "dx", "xunit", "y0", "dy", "yunit"):
            assert getattr(a, attr) == getattr(b, attr)
        utils.assert_array_equal(b[0], a[0])
        utils.assert_array_equal(b.xindex, a.xindex[0:1])
        utils.assert_array_equal(b.yindex, a.yindex)

    def test_two_index_arrays(self):
        """Test that subsetting with two index arrays works correctly."""
        # create an array with indices
        rawa = numpy.arange(12).reshape((4, 3))
        a = Array2D(rawa)
        exp = numpy.array([3, 11])
        ind1, ind2 = numpy.array([1, 3]), numpy.array([0, 2])
        assert_array_equal(rawa[ind1, ind2], exp)
        assert_array_equal(a[ind1, ind2].value, exp)

    def test_is_compatible_yindex(self):
        """Check that irregular arrays are compatible if their yindexes match."""
        y = numpy.logspace(0, 2, num=self.data.shape[1])
        a = self.create(yindex=y)
        b = self.create(yindex=y)
        assert a.is_compatible(b)

    def test_is_compatible_error_yindex(self, array: Array2DType):
        """Check that `Array2D.is_compatible` errors with mismatching indexes."""
        y = numpy.logspace(0, 2, num=self.data.shape[1])
        other = self.create(yindex=y)
        with pytest.raises(ValueError, match="indexes do not match"):
            array.is_compatible(other)

    def test_value_at(self):
        """Test `Array2D.value_at()`."""
        array = self.create()
        assert array.value_at(2, 3) == self.data[2][3] * array.unit
        assert array.value_at(5 * array.xunit, 2 * array.yunit) == (
            self.data[5][2] * array.unit
        )
        with pytest.raises(IndexError):
            array.value_at(1.6, 4.8)

    def test_pad(self):
        """Test `Array2D.pad()` - NOT IMPLEMENTED for Array2D."""
        pytest.skip("not implemented for >1D arrays")

    def test_pad_index(self):
        """Test `Array2D.pad_index()` - NOT IMPLEMENTED for Array2D."""
        pytest.skip("not implemented for >1D arrays")

    def test_pad_asymmetric(self):
        """Test `Array2D.pad_asymmetric()` - NOT IMPLEMENTED for Array2D."""
        pytest.skip("not implemented for >1D arrays")

    def test_single_getitem_not_created(self, array: Array2DType):  # noqa: ARG002
        """Not implemented for Array2D."""
        pytest.skip("not implemented for >1D arrays")



# Stubs for the SWIG-generated `lalframe` module used by GWpy.

# ruff: noqa: N802, N803

import lal

# -- FrameU interface ----------------

class FrameUFrFile: ...

def FrameUFrFileOpen(filename: str, mode: str) -> FrameUFrFile: ...

class FrameUFrameH: ...

def FrameNew(
    epoch: lal.LIGOTimeGPS, duration: float, project: str,
    run: int, frnum: int, detectorFlags: int,
) -> FrameUFrameH: ...

def FrameUFrameHWrite(stream: FrameUFrFile, frame: FrameUFrameH) -> int: ...

class FrameUFrTOC: ...

def FrameUFrTOCRead(frfile: FrameUFrFile) -> FrameUFrTOC: ...
def FrameUFrTOCQueryAdcN(frtoc: FrameUFrTOC) -> int: ...
def FrameUFrTOCQueryAdcName(frtoc: FrameUFrTOC, adc: int) -> str: ...
def FrameUFrTOCQueryDetectorN(frtoc: FrameUFrTOC) -> int: ...
def FrameUFrTOCQueryDetectorName(frtoc: FrameUFrTOC, adc: int) -> str: ...
def FrameUFrTOCQueryDt(frtoc: FrameUFrTOC, pos: int) -> float: ...
def FrameUFrTOCQueryGTimeModf(frtoc: FrameUFrTOC, pos: int) -> tuple[int, float]: ...
def FrameUFrTOCQueryNFrame(frtoc: FrameUFrTOC) -> int: ...
def FrameUFrTOCQueryProcN(frtoc: FrameUFrTOC) -> int: ...
def FrameUFrTOCQueryProcName(frtoc: FrameUFrTOC, proc: int) -> str: ...
def FrameUFrTOCQuerySimN(frtoc: FrameUFrTOC) -> int: ...
def FrameUFrTOCQuerySimName(frtoc: FrameUFrTOC, sim: int) -> str: ...

class FrameUFrChan: ...

def FrameUFrChanQueryName(chan: FrameUFrChan) -> str: ...
def FrameUFrChanQueryTimeOffset(chan: FrameUFrChan) -> float: ...
def FrameUFrChanRead(stream: FrameUFrFile, name: str, pos: int) -> FrameUFrChan: ...

# -- Stream interface --------------

class FrFile: ...

def FrFileQueryChanType(frfile: FrFile, chname: str, pos: int) -> int: ...
def FrFileQueryChanVectorLength(frfile: FrFile, chname: str, pos: int) -> int: ...
def FrFileQueryDt(frfile: FrFile, pos: int) -> float: ...
def FrFileQueryGTime(start: lal.LIGOTimeGPS, frfile: FrFile, pos: int) -> int: ...
def FrFileQueryNFrame(frfile: FrFile) -> int: ...

class FrStream:
    cache: lal.Cache
    epoch: lal.LIGOTimeGPS
    file: FrFile
    fnum: int
    mode: int
    pos: int

def FrStreamCacheOpen(cache: lal.Cache) -> FrStream: ...
def FrStreamOpen(path: str, name: str) -> FrStream: ...
def FrStreamRewind(stream: FrStream) -> None: ...
def FrStreamNext(stream: FrStream) -> None: ...
def FrStreamSeek(stream: FrStream, epoch: lal.LIGOTimeGPS) -> None: ...

def FrStreamGetTimeSeriesType(channel: str, stream: FrStream) -> int: ...

def FrStreamReadINT2TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.INT2TimeSeries: ...

def FrStreamReadINT4TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.INT4TimeSeries: ...

def FrStreamReadINT8TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.INT8TimeSeries: ...

def FrStreamReadUINT2TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.UINT2TimeSeries: ...

def FrStreamReadUINT4TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.UINT4TimeSeries: ...

def FrStreamReadUINT8TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.UINT8TimeSeries: ...

def FrStreamReadREAL4TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.REAL4TimeSeries: ...

def FrStreamReadREAL8TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.REAL8TimeSeries: ...

def FrStreamReadCOMPLEX8TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.COMPLEX8TimeSeries: ...

def FrStreamReadCOMPLEX16TimeSeries(
    stream: FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
    zero: int,
) -> lal.COMPLEX16TimeSeries: ...



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWF I/O utilities for frameCPP."""

from __future__ import annotations

import warnings
from enum import IntEnum
from typing import (
    TYPE_CHECKING,
    overload,
)

import numpy
from LDAStools import frameCPP

from ...segments import Segment
from ...time import (
    LIGOTimeGPS,
    to_gps,
)
from ...utils.enum import NumpyTypeEnum
from ..utils import file_path
from .core import (
    FRDATA_TYPES,
    _series_name,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
    )
    from pathlib import Path
    from typing import (
        IO,
        Literal,
    )

    from ...time import SupportsToGps
    from ...types import Series

_FrVect = frameCPP.FrVect

# -- detectors -----------------------

DETECTORS: dict[str, int] = {
    key[18:]: val for key, val in vars(frameCPP).items()
    if key.startswith("DETECTOR_LOCATION_")
}
DetectorLocation = IntEnum("DetectorLocation", DETECTORS)


# -- type mapping --------------------

class FrVectType(IntEnum, NumpyTypeEnum):
    INT8 = _FrVect.FR_VECT_C
    INT16 = _FrVect.FR_VECT_2S
    INT32 = _FrVect.FR_VECT_4S
    INT64 = _FrVect.FR_VECT_8S
    FLOAT32 = _FrVect.FR_VECT_4R
    FLOAT64 = _FrVect.FR_VECT_8R
    COMPLEX64 = _FrVect.FR_VECT_8C
    COMPLEX128 = _FrVect.FR_VECT_16C
    BYTES = _FrVect.FR_VECT_STRING
    UINT8 = _FrVect.FR_VECT_1U
    UINT16 = _FrVect.FR_VECT_2U
    UINT32 = _FrVect.FR_VECT_4U
    UINT64 = _FrVect.FR_VECT_8U


# -- compression types ---------------

try:
    _FrVect.ZERO_SUPPRESS
except AttributeError:  # python-ldas-tools-framecpp < 3.0.0
    class Compression(IntEnum):
        RAW = _FrVect.RAW
        GZIP = _FrVect.GZIP
        DIFF_GZIP = _FrVect.DIFF_GZIP
        ZERO_SUPPRESS_WORD_2 = _FrVect.ZERO_SUPPRESS_WORD_2
        ZERO_SUPPRESS_WORD_4 = _FrVect.ZERO_SUPPRESS_WORD_4
        ZERO_SUPPRESS_WORD_8 = _FrVect.ZERO_SUPPRESS_WORD_8
        ZERO_SUPPRESS_OTHERWISE_GZIP = _FrVect.ZERO_SUPPRESS_OTHERWISE_GZIP
else:
    class Compression(IntEnum):
        RAW = _FrVect.RAW
        BIGENDIAN_RAW = _FrVect.BIGENDIAN_RAW
        LITTLEENDIAN_RAW = _FrVect.LITTLEENDIAN_RAW
        GZIP = _FrVect.GZIP
        BIGENDIAN_GZIP = _FrVect.BIGENDIAN_GZIP
        LITTLEENDIAN_GZIP = _FrVect.LITTLEENDIAN_GZIP
        DIFF_GZIP = _FrVect.DIFF_GZIP
        BIGENDIAN_DIFF_GZIP = _FrVect.BIGENDIAN_DIFF_GZIP
        LITTLEENDIAN_DIFF_GZIP = _FrVect.LITTLEENDIAN_DIFF_GZIP
        ZERO_SUPPRESS = _FrVect.ZERO_SUPPRESS
        BIGENDIAN_ZERO_SUPPRESS = _FrVect.BIGENDIAN_ZERO_SUPPRESS
        LITTLEENDIAN_ZERO_SUPPRESS = _FrVect.LITTLEENDIAN_ZERO_SUPPRESS
        ZERO_SUPPRESS_OTHERWISE_GZIP = _FrVect.ZERO_SUPPRESS_OTHERWISE_GZIP

# compression level is '6' for all GZip compressions, otherwise 0 (none)
DefaultCompressionLevel = IntEnum(
    "DefaultCompressionLevel",
    {k: 6 if "GZIP" in k else 0 for k in Compression.__members__},
)


# -- Proc data types -----------------

class FrProcDataType(IntEnum):
    UNKNOWN = frameCPP.FrProcData.UNKNOWN_TYPE
    TIME_SERIES = frameCPP.FrProcData.TIME_SERIES
    FREQUENCY_SERIES = frameCPP.FrProcData.FREQUENCY_SERIES
    OTHER_1D_SERIES_DATA = frameCPP.FrProcData.OTHER_1D_SERIES_DATA
    TIME_FREQUENCY = frameCPP.FrProcData.TIME_FREQUENCY
    WAVELETS = frameCPP.FrProcData.WAVELETS
    MULTI_DIMENSIONAL = frameCPP.FrProcData.MULTI_DIMENSIONAL


class FrProcDataSubType(IntEnum):
    UNKNOWN = frameCPP.FrProcData.UNKNOWN_SUB_TYPE
    DFT = frameCPP.FrProcData.DFT
    AMPLITUDE_SPECTRAL_DENSITY = frameCPP.FrProcData.AMPLITUDE_SPECTRAL_DENSITY
    POWER_SPECTRAL_DENSITY = frameCPP.FrProcData.POWER_SPECTRAL_DENSITY
    CROSS_SPECTRAL_DENSITY = frameCPP.FrProcData.CROSS_SPECTRAL_DENSITY
    COHERENCE = frameCPP.FrProcData.COHERENCE
    TRANSFER_FUNCTION = frameCPP.FrProcData.TRANSFER_FUNCTION


# -- I/O functions -------------------


@overload
def open_gwf(
    gwf: str | Path | IO | frameCPP.IFrameFStream | frameCPP.OFrameFStream,
    mode: Literal["r"],
) -> frameCPP.IFrameFStream:
    ...

@overload
def open_gwf(
    gwf: str | Path | IO | frameCPP.IFrameFStream | frameCPP.OFrameFStream,
    mode: Literal["w"],
) -> frameCPP.OFrameFStream:
    ...

def open_gwf(
    gwf: str | Path | IO | frameCPP.IFrameFStream | frameCPP.OFrameFStream,
    mode: Literal["r", "w"] = "r",
) -> frameCPP.IFrameFStream | frameCPP.OFrameFStream:
    """Open a stream for reading or writing GWF format data.

    Parameters
    ----------
    gwf : `str`, `pathlib.Path`, `file`, or open frameCPP stream
        The path to read from, or write to. Already open GWF streams
        are returned unmodified, if the type matches the mode.

    mode : `str`, optional
        The mode with which to open the file, either `r` or `w`.

    Returns
    -------
    stream : `LDAStools.frameCPP.IFrameFStream`, or `LDAStools.frameCPP.OFrameFStream`
        The file stream for reading or writing.
    """
    # check mode
    if mode not in ("r", "w"):
        msg = "mode must be either 'r' or 'w'"
        raise ValueError(msg)
    # match open file stream
    if mode == "r" and isinstance(gwf, frameCPP.IFrameFStream):
        return gwf
    if mode == "w" and isinstance(gwf, frameCPP.OFrameFStream):
        return gwf
    # open a new stream
    filename = file_path(gwf)
    if mode == "r":
        return frameCPP.IFrameFStream(filename)
    return frameCPP.OFrameFStream(filename)


def write_frames(
    gwf: str | Path | IO,
    frames: Iterable[frameCPP.FrameH],
    compression: int | str | None = None,
    compression_level: int | None = None,
):
    """Write a list of frame objects to a file.

    **Requires:** |LDAStools.frameCPP|_

    Parameters
    ----------
    filename : `str`
        path to write into

    frames : `list` of `LDAStools.frameCPP.FrameH`
        list of frames to write into file

    compression : `int`, `str`, optional
        name of compresion algorithm to use, or its endian-appropriate
        ID, choose from

        - ``'RAW'``
        - ``'GZIP'`` (default)
        - ``'DIFF_GZIP'``
        - ``'ZERO_SUPPRESS'``
        - ``'ZERO_SUPPRESS_OTHERWISE_GZIP'``

    compression_level : `int`, optional
        compression level for given method, default is ``6`` for GZIP-based
        methods, otherwise ``0``
    """
    # handle compression arguments
    comp: Compression
    if compression is None:
        comp = Compression.GZIP
    elif isinstance(compression, int):
        comp = Compression(compression)
    else:
        comp = Compression[compression]
    if compression_level is None:
        compression_level = DefaultCompressionLevel[comp.name].value

    # open stream for writing
    stream = open_gwf(gwf, "w")

    # write frames one-by-one
    if isinstance(frames, frameCPP.FrameH):
        frames = [frames]
    for frame in frames:
        stream.WriteFrame(frame, int(comp.value), int(compression_level))
    # stream auto-closes (apparently)


def create_frame(
    time: SupportsToGps = 0,
    duration: float | None = None,
    name: str = "gwpy",
    run: int = -1,
    ifos: Iterable[str] | None = None,
) -> frameCPP.FrameH:
    """Create a new :class:`~LDAStools.frameCPP.FrameH`.

    **Requires:** |LDAStools.frameCPP|_

    Parameters
    ----------
    time : `float`, optional
        Frame start time in GPS seconds.

    duration : `float`, optional
        Frame length in seconds.

    name : `str`, optional
        Name of project or other experiment description.

    run : `int`, optional
        Run number (number < 0 reserved for simulated data); monotonic for
        experimental runs.

    ifos : `list`, optional
        List of interferometer prefices (e.g. ``'L1'``) associated with this
        frame.

    Returns
    -------
    frame : :class:`~LDAStools.frameCPP.FrameH`
        The newly created frame header.
    """
    # create frame
    frame = frameCPP.FrameH()

    # add timing
    gps = to_gps(time)
    gps = frameCPP.GPSTime(gps.gpsSeconds, gps.gpsNanoSeconds)
    frame.SetGTime(gps)
    if duration is not None:
        frame.SetDt(float(duration))

    # add FrDetectors
    for prefix in ifos or []:
        frame.AppendFrDetector(
            frameCPP.GetDetector(DetectorLocation[prefix], gps),
        )

    # add descriptions
    frame.SetName(name)
    frame.SetRun(run)

    return frame


def create_fradcdata(
    series: Series,
    frame_epoch: SupportsToGps = 0,
    channelgroup: int = 0,
    channelid: int = 0,
    nbits: int = 16,
) -> frameCPP.FrAdcData:
    """Create a `~frameCPP.FrAdcData` from a `~gwpy.types.Series`.

    .. note::

       Currently this method is restricted to 1-dimensional arrays.

    Parameters
    ----------
    series : `~gwpy.types.Series`
        the input data array to store

    frame_epoch : `float`, `int`, optional
        the GPS start epoch of the `Frame` that will contain this
        data structure

    Returns
    -------
    frdata : `~frameCPP.FrAdcData`
        the newly created data structure

    Notes
    -----
    See Table 10 (§4.3.2.4) of :dcc:`LIGO-T970130` for more details.
    """
    # assert correct type
    if not series.xunit.is_equivalent("s") or series.ndim != 1:
        msg = "only 1-dimensional timeseries data can be written as FrAdcData"
        raise TypeError(msg)

    frdata = frameCPP.FrAdcData(
        _series_name(series),
        channelgroup,
        channelid,
        nbits,
        (1 / series.dx.to("s")).value,
    )
    frdata.SetTimeOffset(
        float(to_gps(series.x0.value) - to_gps(frame_epoch)),
    )
    return frdata


def _get_series_trange(series: Series) -> float:
    if series.xunit.is_equivalent("s"):
        return abs(series.xspan)
    return 0


def _get_series_frange(series: Series) -> float:
    if series.xunit.is_equivalent("Hz"):  # FrequencySeries
        return abs(series.xspan)
    if series.ndim == 2 and series.yunit.is_equivalent("Hz"):  # Spectrogram
        return abs(series.yspan)
    return 0


def create_frprocdata(
    series: Series,
    frame_epoch: SupportsToGps = 0,
    comment: str | None = None,
    type: int | str | None = None,
    subtype: int | str | None = None,
    trange: float | None = None,
    fshift: float = 0,
    phase: float = 0,
    frange: float | None = None,
    bandwidth: float = 0,
) -> frameCPP.FrProcData:
    """Create a `~frameCPP.FrProcData` from a `~gwpy.types.Series`.

    .. note::

       Currently this method is restricted to 1-dimensional arrays.

    Parameters
    ----------
    series : `~gwpy.types.Series`
        the input data array to store

    frame_epoch : `float`, `int`, optional
        the GPS start epoch of the `Frame` that will contain this
        data structure

    comment : `str`, optional
        comment

    type : `int`, `str`, optional
        type of data object

    subtype : `int`, `str`, optional
        subtype for f-Series

    trange : `float`, optional
        duration of sampled data

    fshift : `float`, optional
        frequency in the original data that corresponds to 0 Hz in the
        heterodyned series

    phase : `float`, optional
        phase of the heterodyning signal at start of dataset

    frange : `float`, optional
        frequency range

    bandwidth : `float, optional
        reoslution bandwidth

    Returns
    -------
    frdata : `~frameCPP.FrAdcData`
        the newly created data structure

    Notes
    -----
    See Table 17 (§4.3.2.11) of :dcc:`LIGO-T970130` for more details.
    """
    # format auxiliary data
    if trange is None:
        trange = _get_series_trange(series)
    if frange is None:
        frange = _get_series_frange(series)

    return frameCPP.FrProcData(
        _series_name(series),
        str(comment or series.name),
        _get_frprocdata_type(series, type),
        _get_frprocdata_subtype(series, subtype),
        float(to_gps(series.x0.value) - to_gps(frame_epoch)),
        trange,
        fshift,
        phase,
        frange,
        bandwidth,
    )


def create_frsimdata(
    series: Series,
    frame_epoch: SupportsToGps = 0,
    comment: str | None = None,
    fshift: float = 0,
    phase: float = 0,
) -> frameCPP.FrSimData:
    """Create a `~frameCPP.FrSimData` from a `~gwpy.types.Series`.

    .. note::

       Currently this method is restricted to 1-dimensional arrays.

    Parameters
    ----------
    series : `~gwpy.types.Series`
        the input data array to store

    frame_epoch : `float`, `int`, optional
        the GPS start epoch of the `Frame` that will contain this
        data structure

    fshift : `float`, optional
        frequency in the original data that corresponds to 0 Hz in the
        heterodyned series

    phase : `float`, optional
        phase of the heterodyning signal at start of dataset

    Returns
    -------
    frdata : `~frameCPP.FrSimData`
        the newly created data structure

    Notes
    -----
    See Table 20 (§4.3.2.14) of :dcc:`LIGO-T970130` for more details.
    """
    # assert correct type
    if not series.xunit.is_equivalent("s"):
        msg = "only timeseries data can be written as FrSimData"
        raise TypeError(msg)

    return frameCPP.FrSimData(
        _series_name(series),
        str(comment or series.name),
        (1 / series.dx.to("s")).value,
        float(to_gps(series.x0.value) - to_gps(frame_epoch)),
        fshift,
        phase,
    )


def create_frvect(series: Series) -> frameCPP.FrVect:
    """Create a `~frameCPP.FrVect` from a `~gwpy.types.Series`.

    .. note::

       Currently this method is restricted to 1-dimensional arrays.

    Parameters
    ----------
    series : `~gwpy.types.Series`
        The input data array to store.

    Returns
    -------
    frvect : :class:`LDASTools.frameCPP.FrVect`
        The newly created data vector.
    """
    # create dimensions
    dims = frameCPP.Dimension(
        series.shape[0],  # num elements
        series.dx.value,  # step size
        str(series.dx.unit),  # unit
        0,  # starting value
    )

    # create FrVect
    vect = frameCPP.FrVect(
        _series_name(series),  # name
        int(FrVectType.find(series.dtype).value),  # data type enum
        series.ndim,  # num dimensions
        dims,  # dimension object
        str(series.unit),  # unit
    )

    # populate FrVect
    vect.GetDataArray()[:] = numpy.require(series.value, requirements=["C"])

    return vect


# -- utilities -----------------------

@overload
def _iter_toc(
    gwf: str | Path | IO | frameCPP.IFrameFStream,
    type: str | None,
    count: Literal[True],
) -> Iterator[int]:
    ...

@overload
def _iter_toc(
    gwf: str | Path | IO | frameCPP.IFrameFStream,
    type: str | None,
    count: Literal[False],
) -> Iterator[tuple[str, str]]:
    ...

def _iter_toc(
    gwf: str | Path | IO | frameCPP.IFrameFStream,
    type: str | None = None,
    count: bool = False,
) -> Iterator[tuple[str, str] | int]:
    """Yields the name and type of each channel in a GWF file TOC.

    **Requires:** |LDAStools.frameCPP|_

    Parameters
    ----------
    gwf : `str`, `LDAStools.frameCPP.IFrameFStream`
        Path of GWF file, or open file stream, to read.

    type : `str`, optional
        Only yield items of the given type (case insensitive).
        Default is all types 'adc', 'proc', 'or 'sim'.

    Yields
    ------
    name, type : `str`, `str`
        The data channel name and type (one of 'adc', 'proc', or 'sim').
    """
    stream = open_gwf(gwf, mode="r")
    toc = stream.GetTOC()
    if type is None:
        types = FRDATA_TYPES
    else:
        types = (type,)
    for typename in map(str.lower, types):
        if typename == "adc":
            getter = toc.GetADC
        else:
            getter = getattr(toc, f"Get{typename.title()}")
        if count:  # just count, don't unpack
            yield len(getter())
            continue
        for name in getter():
            yield name, typename


def _count_toc(
    gwf: str | Path | IO | frameCPP.IFrameFStream,
    type: str | None = None,
) -> int:
    """Yield the names and types of channels listed in the TOC for a GWF file.

    Parameters
    ----------
    gwf : `str`, `pathlib.Path`, `file`, `lalframe.FrameUFrFile`
        Path of GWF file, or open file stream, to read.

    Yields
    ------
    name : `str`
        The name of a channel
    type : `str`
        The ``FrVect`` type of the channel, one of ``"sim"``, ``"proc"``,
        or ``"adc"``.
    """
    return sum(_iter_toc(gwf, type=type, count=True))


def _channel_segments(
    gwf: str | Path | IO | frameCPP.IFrameFStream,
    channel: str,
    warn: bool = True,
) -> Iterator[Segment]:
    """Yields the segments containing data for ``channel`` in this GWF path."""
    stream = open_gwf(gwf, mode="r")

    # get segments for frames
    toc = stream.GetTOC()
    secs = toc.GetGTimeS()
    nano = toc.GetGTimeN()
    dur = toc.GetDt()

    readers: tuple[Callable, ...] = tuple(
        getattr(stream, f"ReadFr{type_.title()}Data")
        for type_ in FRDATA_TYPES
    )

    # for each segment, try and read the data for this channel
    for i, (s, ns, dt) in enumerate(zip(secs, nano, dur, strict=True)):
        for read in readers:
            try:
                read(i, channel)
            except (IndexError, ValueError):
                continue
            readers = (read,)  # use this one from now on
            epoch = LIGOTimeGPS(s, ns)
            yield Segment(epoch, epoch + dt)
            break
        else:  # none of the readers worked for this channel, warn
            if warn:
                warnings.warn(
                    f"'{channel}' not found in frame {i} of {gwf}",
                    stacklevel=2,
                )


def _get_type(
    type_: int | str,
    enum: type[IntEnum],
) -> int:
    """Handle a type string, or just return an `int`.

    Only to be called in relation to FrProcDataType and FrProcDataSubType
    """
    if isinstance(type_, int):
        return type_
    return enum[str(type_).upper()].value


def _get_frprocdata_type(
    series: Series,
    type_: int | str | None,
) -> int:
    """Determine the appropriate `FrProcDataType` for this series.

    Notes
    -----
    See Table 17 (§4.3.2.11) of :dcc:`LIGO-T970130` for more details.
    """
    if type_ is not None:  # format user value
        return _get_type(type_, FrProcDataType)

    if series.ndim == 1:
        if series.xunit.is_equivalent("s"):
            return FrProcDataType.TIME_SERIES.value
        if series.xunit.is_equivalent("Hz"):
            return FrProcDataType.FREQUENCY_SERIES.value
        return FrProcDataType.OTHER_1D_SERIES_DATA.value
    if (
        series.ndim == 2
        and series.xunit.is_equivalent("s")
        and series.yunit.is_equivalent("Hz")
    ):
        return FrProcDataType.TIME_FREQUENCY.value
    if series.ndim > 2:
        return FrProcDataType.MULTI_DIMENSIONAL.value

    return FrProcDataType.UNKNOWN.value


def _get_frprocdata_subtype(
    series: Series,
    subtype: int | str | None,
) -> int:
    """Determine the appropriate `FrProcDataSubType` for this series.

    Notes
    -----
    See Table 17 (§4.3.2.11) of :dcc:`LIGO-T970130` for more details.
    """
    if subtype is not None:  # format user value
        return _get_type(subtype, FrProcDataSubType)

    if series.unit == "coherence":
        return FrProcDataSubType.COHERENCE.value
    return FrProcDataSubType.UNKNOWN.value



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.units`."""

import pytest
from astropy.units import Unit

from .. import units as plot_units


class TestLatexInlineDimensional:
    """Test for `gwpy.plot.units.LatexInlineDimensional`."""

    Format = plot_units.LatexInlineDimensional

    @pytest.mark.parametrize(("unit", "out"), [
        pytest.param(
            Unit("m"),
            r"Length [$\mathrm{m}$]",
            id="length",
        ),
        pytest.param(
            Unit("m/s"),
            r"Speed/Velocity [$\mathrm{m\,s^{-1}}$]",
            id="speed",
        ),
        pytest.param(
            Unit(""),
            "Dimensionless",
            id="dimensionless",
        ),
        pytest.param(
            Unit("strain"),
            r"[$\mathrm{strain}$]",
            id="strain",
        ),
    ])
    def test_to_string(self, unit, out):
        """Test `to_string()`."""
        assert unit.to_string(self.Format.name) == out



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O utilities for ROOT files."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal

    from .utils import (
        FileLike,
        FileSystemPath,
    )

__all__ = [
    "identify_root",
]

ROOT_SIGNATURE = b"root"
ROOT_SIGNATURE_T = ROOT_SIGNATURE.decode("utf-8")


def identify_root(
    origin: Literal["read", "write"],  # noqa: ARG001
    filepath: FileSystemPath | None,
    fileobj: FileLike | None,
    *args,  # noqa: ANN002
    **kwargs,  # noqa: ARG001
) -> bool:
    """Identify a filename or file object as ROOT.

    This function is only indented to be used as a registered
    identifier for Astropy's Unified I/O system.
    """
    # try and read file descriptor
    if fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            signature = fileobj.read(4)
        finally:
            fileobj.seek(loc)
        return signature in (ROOT_SIGNATURE, ROOT_SIGNATURE_T)
    # otherwise guess from file name
    if filepath is not None:
        return str(filepath).endswith(".root")
    # or finally introspect an object from uproot
    try:
        import uproot
    except ImportError:
        return False
    return isinstance(args[0], uproot.ReadOnlyFile | uproot.ReadOnlyDirectory)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting utilities for segments."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

import igwn_segments
from matplotlib.artist import allow_rasterization
from matplotlib.collections import PatchCollection
from matplotlib.colors import is_color_like
from matplotlib.patches import Rectangle
from matplotlib.projections import register_projection
from matplotlib.ticker import (
    Formatter,
    MultipleLocator,
)

from .axes import Axes
from .colors import tint
from .text import to_string

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Literal

    from igwn_segments import (
        segment,
        segmentlist,
        segmentlistdict,
    )
    from matplotlib.artists import Artist
    from matplotlib.backend_bases import RendererBase
    from matplotlib.collections import Collection

    from gwpy.segments import (
        DataQualityDict,
        DataQualityFlag,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

HATCHES = ["/", "\\", "|", "-", "+", "x", "o", "O", ".", "*"]


class SegmentAxes(Axes):
    """Custom `Axes` to display segments.

    This `SegmentAxes` provides custom methods for displaying any of

    - `~gwpy.segments.DataQualityFlag`
    - `~gwpy.segments.Segment` or :class:`igwn_segments.segment`
    - `~gwpy.segments.SegmentList` or :class:`igwn_segments.segmentlist`
    - `~gwpy.segments.SegmentListDict` or
      :class:`igwn_segments.segmentlistdict`

    Parameters
    ----------
    insetlabels : `bool`, default: `False`
        display segment labels inside the axes. Prevents very long segment
        names from getting squeezed off the end of a standard figure

    See Also
    --------
    gwpy.plot.Axes
        for documentation of other args and kwargs
    """

    name = "segments"

    def __init__(
        self,
        *args,
        xscale: str = "auto-gps",
        insetlabels: bool = False,
        **kwargs,
    ) -> None:
        """Initialise a new `SegmentAxes`."""
        # make axes
        super().__init__(
            *args,
            xscale=xscale,
            insetlabels=insetlabels,
            **kwargs,
        )

        # set y-axis labels
        self.yaxis.set_major_locator(MultipleLocator())
        formatter = SegmentFormatter()
        self.yaxis.set_major_formatter(formatter)

    def _plot_method(
        self,
        obj: DataQualityFlag | DataQualityDict | segmentlist | segmentlistdict,
    ) -> Callable:
        from ..segments import (
            DataQualityDict,
            DataQualityFlag,
        )

        if isinstance(obj, DataQualityDict):
            return self.plot_dict
        if isinstance(obj, DataQualityFlag):
            return self.plot_flag
        if isinstance(obj, igwn_segments.segmentlistdict):
            return self.plot_segmentlistdict
        if isinstance(obj, igwn_segments.segmentlist):
            return self.plot_segmentlist
        msg = (
            f"no known {type(self).__name__}.plot_xxx method "
            f"for {type(obj).__name__}"
        )
        raise TypeError(msg)

    def plot(
        self,
        *args: DataQualityFlag | DataQualityDict | segmentlist | segmentlistdict,
        **kwargs,
    ) -> list[Artist]:
        """Plot data onto these axes.

        Parameters
        ----------
        args
            A single instance of

                - `~gwpy.segments.DataQualityFlag`
                - `~gwpy.segments.Segment`
                - `~gwpy.segments.SegmentList`
                - `~gwpy.segments.SegmentListDict`

            or equivalent types upstream from :mod:`igwn_segments`.

        kwargs
            All keyword arguments are passed to `~matplotib.axes.Axes.plot`
            when drawing each object.

        Returns
        -------
        artists : `list` of `~matplotlib.artist.Artist`
            The list of things that were rendered.

        See Also
        --------
        matplotlib.axes.Axes.plot
            for a full description of acceptable ``*args` and ``**kwargs``.
        """
        out = []

        # loop over the various arguments, and plot them
        args = list(args)
        while args:
            try:
                plotter = self._plot_method(args[0])
            except TypeError:
                break
            out.append(plotter(args[0], **kwargs))
            args.pop(0)

        if args:
            out.extend(super().plot(*args, **kwargs))

        self.autoscale(
            enable=None,
            axis="both",
            tight=False,
        )
        return out

    def plot_dict(
        self,
        flags: DataQualityDict,
        label: str = "key",
        known: str | dict | None = "x",
        **kwargs,
    ) -> list[PatchCollection]:
        """Plot a `~gwpy.segments.DataQualityDict` onto these axes.

        Parameters
        ----------
        flags : `~gwpy.segments.DataQualityDict`
            The data-quality dict to display.

        label : `str`, optional
            Labelling system to use, or fixed label for all `DataQualityFlags`.
            Special values include

            ``'key'``
                Use the key of the `DataQualityDict`.

            ``'name'``
                Use the :attr:`~DataQualityFlag.name` of the `DataQualityFlag`.

            If anything else, that fixed label will be used for all lines.

        known : `str`, `dict`, `None`, optional
            Display `known` segments with the given hatching, or give a
            dict of keyword arguments to pass to
            :meth:`~SegmentAxes.plot_segmentlist`, or `None` to hide.

        kwargs
            All other keyword arguments are passed to `SegmentRectangle`.

        Returns
        -------
        artists : `list` of `~matplotlib.patches.PatchCollection`
            The list of patch collections that were drawn.
        """
        out = []
        for name, flag in flags.items():
            if label.lower() == "name":
                lab = flag.name
            elif label.lower() == "key":
                lab = name
            else:
                lab = label
            out.append(self.plot_flag(
                flag,
                label=to_string(lab),
                known=known,
                **kwargs,
            ))
        return out

    def plot_flag(
        self,
        flag: DataQualityFlag,
        y: float | None = None,
        **kwargs,
    ) -> PatchCollection | list[SegmentRectangle]:
        """Plot a `~gwpy.segments.DataQualityFlag` onto these axes.

        Parameters
        ----------
        flag : `~gwpy.segments.DataQualityFlag`
            Data-quality flag to display.

        y : `float`, optional
            Y-axis value for new segments.

        height : `float`, optional,
            Height for each segment, default: `0.8`.

        known : `str`, `dict`, `None`
            One of the following

            - ``'fancy'`` - to use fancy format (try it and see)
            - ``'x'`` (or similar) - to use hatching
            - `str` to specify ``facecolor`` for known segmentlist
            - `dict` of kwargs to use
            - `None` to ignore known segmentlist

        kwargs
            All other keyword arguments are passed to `SegmentRectangle`.

        Returns
        -------
        artists : `list` of `~matplotlib.patches.PatchCollection`
            The list of patch collections that were drawn.
        """
        # get y axis position
        if y is None:
            y = self.get_next_y()

        # default a 'good' flag to green segments and vice-versa
        if flag.isgood:
            kwargs.setdefault("facecolor", "#33cc33")
            kwargs.setdefault("known", "#ff0000")
        else:
            kwargs.setdefault("facecolor", "#ff0000")
            kwargs.setdefault("known", "#33cc33")
        known = kwargs.pop("known")

        # get flag name
        name = kwargs.pop("label", flag.label or flag.name)

        # make active collection
        kwargs.setdefault("zorder", 0)
        coll = self.plot_segmentlist(
            flag.active,
            y=y,
            label=name,
            **kwargs,
        )

        # make known collection
        if known not in (None, False):
            known_kw = {
                "facecolor": coll.get_facecolor()[0],
                "collection": "ignore",
                "zorder": -1000,
            }
            if isinstance(known, dict):
                known_kw.update(known)
            elif known == "fancy":
                known_kw.update(height=kwargs.get("height", .8) * .05)
            elif known in HATCHES:
                known_kw.update(fill=False, hatch=known)
            else:
                known_kw.update(
                    fill=True,
                    facecolor=known,
                    height=kwargs.get("height", .8) * .5,
                )
            self.plot_segmentlist(flag.known, y=y, label=name, **known_kw)

        return coll

    def plot_segmentlist(
        self,
        segmentlist: segmentlist,
        y: float | None = None,
        height: float = .8,
        label: str | None = None,
        *,
        collection: bool = True,
        rasterized: bool | None = None,
        **kwargs,
    ) -> PatchCollection | list[SegmentRectangle]:
        """Plot a `~gwpy.segments.SegmentList` onto these axes.

        Parameters
        ----------
        segmentlist : `~gwpy.segments.SegmentList`
            List of segments to display.

        y : `float`, optional
            Y-axis value for new segments.

        height : `float`, optional
            Height (in y-axis units) for segment.

        collection : `bool`, optional
            If `True` (default), bundle all patches as a
            `~matplotlib.collections.PatchCollection`, otherwise
            just return a `list` of `SegmentRectangle` patches.

        rasterized : `bool`, optional
            If `True`, rasterise the patches when drawing.
            Default is `False`.

        label : `str`, optional
            custom descriptive name to print as y-axis tick label

        kwargs
            All other keyword arguments are passed to `SegmentRectangle`.

        Returns
        -------
        patches : `~matplotlib.patches.PatchCollection` or `list[SegmentRectangle]`
            The drawn patches, bundled in a `PatchCollection` if ``collection=True``.
        """
        # get colour
        facecolor = kwargs.pop("facecolor", kwargs.pop("color", "#629fca"))
        if is_color_like(facecolor):
            kwargs.setdefault("edgecolor", tint(facecolor, factor=.5))

        # get y
        if y is None:
            y = self.get_next_y()

        # build patches
        patches = [
            SegmentRectangle(
                seg,
                y,
                height=height,
                facecolor=facecolor,
                **kwargs,
            )
            for seg in segmentlist
        ]

        if collection:  # map to PatchCollection
            coll = PatchCollection(
                patches,
                match_original=patches,
                zorder=kwargs.get("zorder", 1),
            )
            coll.set_rasterized(rasterized)
            coll._ignore = collection == "ignore"
            coll._ypos = y
            out = self.add_collection(coll)
            # reset label with tex-formatting now
            #   matplotlib default label is applied by add_collection
            #   so we can only replace the leading underscore after
            #   this point
            if label is None:
                label = coll.get_label()
            coll.set_label(to_string(label))
        else:
            out = []
            for patch in patches:
                patch.set_label(label)
                patch.set_rasterized(rasterized)
                label = ""  # don't label any other patches
                out.append(self.add_patch(patch))
        self.autoscale(enable=None, axis="both", tight=False)
        return out

    def plot_segmentlistdict(
        self,
        segmentlistdict: segmentlistdict,
        y: float | None = None,
        dy: float = 1,
        **kwargs,
    ) -> PatchCollection | list[SegmentRectangle]:
        """Plot a `~gwpy.segments.SegmentListDict` onto these axes.

        Parameters
        ----------
        segmentlistdict : `~gwpy.segments.SegmentListDict`
            (name, `~gwpy.segments.SegmentList`) dict.

        y : `float`, optional
            Starting y-axis value for new segmentlists.

        dy : `float`, optional
            Y-axis separation between each (anchor of each) segmentlist.

        kwargs
            All other keyword arguments are passed to
            `~matplotlib.patches.Rectangle`.

        Returns
        -------
        collections : `list` of `PatchCollection`
            List of `~matplotlib.patches.PatchCollection` sets for
            each segmentlist.
        """
        if y is None:
            y = self.get_next_y()
        collections = []
        for name, segmentlist in segmentlistdict.items():
            collections.append(self.plot_segmentlist(
                segmentlist,
                y=y,
                label=name,
                **kwargs,
            ))
            y += dy
        return collections

    def get_next_y(self) -> int:
        """Find the next y-axis value at which a segment list can be placed.

        This method simply counts the number of independent segmentlists or
        flags that have been plotted onto these axes.
        """
        return len(self.get_collections(ignore=False))

    def get_collections(self, ignore: bool | None = None) -> list[Collection]:
        """Return the collections matching the given `_ignore` value.

        Parameters
        ----------
        ignore : `bool`, or `None`
            Value of `_ignore` to match.

        Returns
        -------
        collections : `list` of `~matplotlib.collections.Collection`
            If `ignore=None`, simply returns all collections, otherwise
            returns those collections matching the `ignore` parameter.
        """
        if ignore is None:
            return self.collections
        return [
            c for c in self.collections
            if getattr(c, "_ignore", None) == ignore
        ]

    def set_insetlabels(self, inset: bool | None = None) -> None:
        """Set the labels to be inset or not.

        Parameters
        ----------
        inset : `bool`, `None`
            Enable (`True`) or disable (`False`) inset labels.
            Default (`None`) toggles the current state.
        """
        self._insetlabels = not self._insetlabels if inset is None else inset

    def get_insetlabels(self) -> bool:
        """Return the inset labels state."""
        return self._insetlabels

    insetlabels = property(
        fget=get_insetlabels,
        fset=set_insetlabels,
        doc=get_insetlabels.__doc__,
    )

    @allow_rasterization
    @wraps(Axes.draw)
    def draw(self, renderer: RendererBase) -> None:
        """Draw the current `SegmentAxes`."""
        # inset the labels if requested
        for tick in self.get_yaxis().get_ticklabels():
            if self.get_insetlabels():
                # record parameters we are changing
                tick._orig_bbox = tick.get_bbox_patch()
                tick._orig_ha = tick.get_ha()
                tick._orig_pos = tick.get_position()
                # modify tick
                tick.set_horizontalalignment("left")
                tick.set_position((0.01, tick.get_position()[1]))
                tick.set_bbox({
                    "alpha": 0.5,
                    "edgecolor": "none",
                    "facecolor": "white",
                })
            elif self.get_insetlabels() is False:
                # if label has been moved, reset things
                try:
                    tick.set_bbox(tick._orig_bbox)
                except AttributeError:
                    pass
                else:
                    tick.set_horizontalalignment(tick._orig_ha)
                    tick.set_position(tick._orig_pos)
                    del tick._orig_bbox
                    del tick._orig_ha
                    del tick._orig_pos
        return super().draw(renderer)

    draw.__doc__ = Axes.draw.__doc__


register_projection(SegmentAxes)


class SegmentFormatter(Formatter):
    """Custom tick formatter for y-axis flag names."""

    def __call__(
        self,
        t: float,
        pos: float | None = None,  # noqa: ARG002
    ) -> str:
        """Format ticks using segment names."""
        ax = self.axis.axes

        # if segments have been plotted at this y-axis value, continue
        for coll in ax.get_collections(ignore=False):
            if t == getattr(coll, "_ypos", None):
                return coll.get_label()

        for patch in ax.patches:
            # if this patch doesn't have a label or doesn't want one, carry on
            if not patch.get_label() or patch.get_label() == "_nolegend_":
                continue
            # otherwise if the axis position overlaps the bbox, emit the
            # patch label as the tick label
            if t in igwn_segments.segment(*patch.get_bbox().intervaly):
                return patch.get_label()

        return ""


# -- segment patch -------------------

class SegmentRectangle(Rectangle):
    """Custom `~matplotlib.patches.Rectangle` for a `~gwpy.segments.Segment`."""

    def __init__(
        self,
        segment: segment,
        y: float,
        height: float = .8,
        valign: Literal["top", "center", "bottom"] = "center",
        **kwargs,
    ) -> None:
        """Build a `~matplotlib.patches.Rectangle` from a segment.

        Parameters
        ----------
        segment : `~gwpy.segments.Segment`
            ``[start, stop)`` GPS segment.

        y : `float`
            Y-axis position for segment.

        height : `float`, optional
            Height (in y-axis units) for segment.

        valign : `str`, optional
            Alignment of segment on y-axis value, one of
            ``"top"``, ``"center"``, or ``"bottom"``.
            Default: ``"center"``.

        kwargs
            All other keyword arguments are passed to
            `~matplotlib.patches.Rectangle`.

        Raises
        ------
        ValueError
            If an invalid ``valign`` value is given.
        """
        if valign.lower() == "bottom":
            y0 = y
        elif valign.lower() in {"center", "centre"}:
            y0 = y - height / 2.
        elif valign.lower() == "top":
            y0 = y - height
        else:
            msg = "valign must be one of 'top', 'center', or 'bottom'"
            raise ValueError(msg)
        width = segment[1] - segment[0]

        super().__init__(
            (segment[0], y0),
            width=width,
            height=height,
            **kwargs,
        )
        self.segment = segment



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. currentmodule:: gwpy.timeseries
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>

Calculate the coherence between two channels
############################################

The `coherence <http://en.wikipedia.org/wiki/Coherence_(physics)>`_ between
two channels is a measure of the frequency-domain correlation between their
time-series data.

In LIGO, the coherence is a crucial indicator of how noise sources couple into
the main differential arm-length readout.
Here we use use the :meth:`TimeSeries.coherence` method to highlight coupling
of the mains power supply into the strain output of the LIGO-Hanford
interferometer.

These data are available as part of the |GWOSC_O3_AUX_RELEASE|_.
"""

# %%
# Data access
# -----------
# First, we import the `TimeSeries`

from gwpy.timeseries import TimeSeries

# %%
# and then :meth:`~TimeSeries.get` the LIGO-Hanford strain data,
# and the mains power monitor data, for a 600-second window near the
# end of the |GWOSC_O3_AUX_RELEASE|_:

start = 1389410000
end = 1389410600
strain = TimeSeries.get("H1", start, end)
mains = TimeSeries.get(
    "H1:PEM-EY_MAINSMON_EBAY_1_DQ",
    start,
    end,
    frametype="H1_AUX_AR1",
)

# %%
#
# .. admonition:: Data are accessed separately
#     :class: note
#
#     We could have used `TimeSeriesDict.get` to access both channels in a
#     single call - that method would first try to find a single dataset that
#     contain both of them, then automatically fall back to separate calls if
#     that fails.
#
#     But, since we know that these channels are in different datsets, we
#     access them separately here for clarity and speed.
#
# Calculating coherence
# ---------------------
# We can then calculate the :meth:`~TimeSeries.coherence` of one
# `TimeSeries` with respect to the other, using an 8-second Fourier
# transform length, with a 4-second (50%) overlap:

coh = strain.coherence(mains, fftlength=8, overlap=4)

# %%
#
# The output of this method is a :class:`~gwpy.frequencyseries.FrequencySeries`
# containing the coherence values for each frequency bin.
#
# Visualisation
# -------------
# We can now :meth:`~gwpy.frequencyseries.FrequencySeries.plot` the coherence:

plot = coh.plot(
    xlabel="Frequency [Hz]",
    xscale="log",
    ylabel="Coherence",
    yscale="linear",
    ylim=(0, 1),
)
plot.show()

# %%
# Here we can see strong coherence at 60 Hz and its harmonics, indicating
# that the mains power supply is coupling into the differential arm length
# control loop.



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from an Omega-format ASCII file."""

from __future__ import annotations

import re

from astropy.io.ascii import core
from astropy.io.registry import (
    get_reader as astropy_get_reader,
    get_writer as astropy_get_writer,
)

from .. import (
    EventTable,
    Table,
)
from .utils import read_with_columns_and_where


class OmegaHeader(core.BaseHeader):
    """Parser for Omega ASCII header."""

    def get_cols(self, lines: list[str]) -> None:
        """Initialize Column objects from a multi-line ASCII header.

        Parameters
        ----------
        lines : `list` or `str`
            List of table lines.
        """
        re_name_def = re.compile(r"^\s*%\s+(?P<colname>\w+)")
        self.names = []
        for line in lines:
            if not line:  # ignore empty lines in header (windows)
                continue
            if not line.startswith("%"):  # end of header lines
                break
            match = re_name_def.search(line)
            if match:
                self.names.append(match.group("colname"))

        if not self.names:
            msg = "No column names found in Omega header"
            raise core.InconsistentTableError(msg)

        self.cols = []
        for name in self.names:
            col = core.Column(name=name)
            self.cols.append(col)

    def write(self, lines: list[str]) -> None:
        """Write column headers to ASCII lines."""
        lines.extend(f"% {name}" for name in self.colnames)


class OmegaData(core.BaseData):
    """Parser for Omega ASCII data."""

    comment = "%"


class Omega(core.BaseReader):
    """Read an Omega file."""

    _format_name = "omega"
    _io_registry_can_write = True
    _description = "Omega format table"

    header_class = OmegaHeader
    data_class = OmegaData


# register for EventTable
EventTable.read.registry.register_reader(
    "ascii.omega",
    EventTable,
    read_with_columns_and_where(astropy_get_reader("ascii.omega", Table)),
)
EventTable.write.registry.register_writer(
    "ascii.omega",
    EventTable,
    astropy_get_writer("ascii.omega", Table),
)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""HDF5 input/output methods for the TimeSeries."""

from __future__ import annotations

from typing import TYPE_CHECKING

from astropy import units

from ...io.hdf5 import (
    identify_hdf5,
    with_read_hdf5,
    with_write_hdf5,
)
from ...types.io.hdf5 import (
    read_hdf5_array,
    write_hdf5_series,
)
from .. import (
    StateVector,
    StateVectorDict,
    TimeSeries,
    TimeSeriesDict,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
    )

    import h5py

SEC_UNIT = units.second

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

def read_hdf5_timeseries(
    h5f: h5py.Group,
    path: str | None = None,
    start: float | None = None,
    end: float | None = None,
    **kwargs,
) -> TimeSeries:
    """Read a `TimeSeries` from HDF5."""
    # read data
    kwargs.setdefault("array_type", TimeSeries)
    series = read_hdf5_array(h5f, path=path, **kwargs)
    # crop if needed
    if start is not None:
        start = max(start, series.span[0])
    if end is not None:
        end = min(end, series.span[1])
    if start is not None or end is not None:
        return series.crop(start, end)
    return series


def _is_timeseries_dataset(dataset: h5py.Dataset) -> bool:
    """Return `True` if a dataset contains `TimeSeries` data."""
    return SEC_UNIT.is_equivalent(dataset.attrs.get("xunit", "undef"))


@with_read_hdf5
def read_hdf5_dict(
    h5f: h5py.Group,
    names: Iterable[str] | None = None,
    group: str | None = None,
    **kwargs,
) -> dict:
    """Read a `TimeSeriesDict` from HDF5."""
    # find group from which to read
    if group:
        h5g = h5f[group]
    else:
        h5g = h5f

    # find list of names to read
    if names is None:
        names = [key for key in h5g if _is_timeseries_dataset(h5g[key])]

    # read names
    out = kwargs.pop("dict_type", TimeSeriesDict)()
    kwargs.setdefault("array_type", out.EntryClass)
    for name in names:
        out[name] = read_hdf5_timeseries(h5g[name], **kwargs)

    return out


def read_hdf5_factory(data_class: type) -> Callable:
    """Create a function to read HDF5 data for the given class type."""
    if issubclass(data_class, dict):
        def read_(*args, **kwargs):  # noqa: ANN002,ANN202
            kwargs.setdefault("dict_type", data_class)
            return read_hdf5_dict(*args, **kwargs)
    else:
        def read_(*args, **kwargs):  # noqa: ANN002,ANN202
            kwargs.setdefault("array_type", data_class)
            return read_hdf5_timeseries(*args, **kwargs)

    return read_


# -- write ---------------------------

@with_write_hdf5
def write_hdf5_dict(
    tsdict: dict,
    h5f: h5py.Group,
    group: str | None = None,
    **kwargs,
) -> None:
    """Write a `TimeSeriesBaseDict` to HDF5.

    Each series in the dict is written as a dataset in the group
    """
    # create group if needed
    if group and group not in h5f:
        h5g = h5f.create_group(group)
    elif group:
        h5g = h5f[group]
    else:
        h5g = h5f

    # write each timeseries
    kwargs.setdefault("format", "hdf5")
    for key, series in tsdict.items():
        series.write(h5g, path=str(key), **kwargs)


# -- register ------------------------

# series classes
for series_class in (TimeSeries, StateVector):
    registry = series_class.read.registry
    reader = read_hdf5_factory(series_class)
    registry.register_reader("hdf5", series_class, reader)
    registry.register_writer("hdf5", series_class, write_hdf5_series)
    registry.register_identifier("hdf5", series_class, identify_hdf5)

# dict classes
for dict_class in (TimeSeriesDict, StateVectorDict):
    registry = series_class.read.registry
    reader = read_hdf5_factory(dict_class)
    registry.register_reader("hdf5", dict_class, reader)
    registry.register_writer("hdf5", dict_class, write_hdf5_dict)
    registry.register_identifier("hdf5", dict_class, identify_hdf5)
